import json
from urllib.parse import urlparse, parse_qs
from pathlib import Path
import re
import os

import pytest
import responses
from responses import matchers

from pytos2.securetrack import StAPI
from pytos2.securetrack.entrypoint import St


@pytest.fixture
def st_api():
    return StAPI(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def st():
    return St(username="username", password="password", hostname="198.18.0.1")


@pytest.fixture
def st_no_cache():
    return St(
        username="username",
        password="password",
        hostname="198.18.0.1",
        cache=False,
        default=False,
    )


@pytest.fixture
def generic_devices_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/generic_devices",
        json=json.load(
            open("tests/securetrack/json/generic_devices/generic_devices.json")
        ),
    )

    (
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/generic_devices?name=vm",
            json=json.load(
                open(
                    "tests/securetrack/json/generic_devices/generic_devices_filtered.json"
                )
            ),
            match_querystring=True,
        ),
    )

    responses.add(
        responses.POST, "https://198.18.0.1/securetrack/api/generic_devices", status=201
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securetrack/api/generic_devices/2",
        status=204,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/generic_devices/3?update_topology=False",
        status=204,
    )


@pytest.fixture
def generic_devices_getter_error(st):
    (
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/generic_devices",
            status=500,
        ),
    )
    pass


@pytest.fixture
def generic_devices_error_mock(st):
    (
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/generic_devices",
            json=json.load(
                open("tests/securetrack/json/generic_devices/generic_devices.json")
            ),
        ),
    )

    responses.add(
        responses.POST, "https://198.18.0.1/securetrack/api/generic_devices", status=400
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securetrack/api/generic_devices/2",
        status=400,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/generic_devices/3?update_topology=False",
        status=400,
    )


@pytest.fixture
def sample_generic_device_csv():
    return open(
        "tests/securetrack/json/generic_devices/sample_generic_device.csv"
    ).read()


@pytest.fixture
def topology_sync_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/topology/synchronize",
        status=200,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/synchronize/status",
        status=200,
        json=json.load(open("tests/securetrack/json/topology/status.json")),
    )


@pytest.fixture
def topology_sync_auth_error_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/topology/synchronize",
        status=401,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/synchronize/status",
        status=401,
    )


@pytest.fixture
def topology_sync_500_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/topology/synchronize",
        status=500,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/synchronize/status",
        status=500,
    )


@pytest.fixture
def topology_sync_502_mock():
    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/topology/synchronize",
        status=502,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/synchronize/status",
        status=502,
    )


@pytest.fixture
def devices_mock(st):
    def device_callback(request):
        parts = request.url.split("/")
        if "show" in parts[-1]:
            _id = parts[-1].split("?", 1)[0]
        else:
            _id = int(parts[-1])

        try:
            _json = json.load(open(f"tests/securetrack/json/devices/device-{_id}.json"))
        except FileNotFoundError:
            return (404, {}, "{}")

        _json = {"device": _json}
        return (200, {}, json.dumps(_json))

    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/devices/(\\d+)$"),
        callback=device_callback,
    )
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/devices/(\\d+)\?show_license=false&show_os_version=false$"
        ),
        callback=device_callback,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices",
        json=json.load(open("tests/securetrack/json/devices/devices.json")),
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/devices/bulk/delete",
        json={"task_uid": "some_guid_here"},
        status=200,
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/devices/bulk/update_topology_data",
        json={"task_uid": "some_guid_here"},
        status=201,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/bulk/tasks/ada853b8-46f7-474b-bb4e-3309a3a9d0af",
        json=json.load(
            open("tests/securetrack/json/devices/get_bulk_device_task.json")
        ),
        status=200,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/bulk/tasks/c878cb8c-6a6d-4939-b20b-550def656ac4",
        json=json.load(
            open("tests/securetrack/json/devices/devices_bulk_delete_task.json")
        ),
    )


@pytest.fixture
def device_rules_mock(st):
    def device_rules_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(open(f"tests/securetrack/json/rules/device-{_id}.json"))
        except FileNotFoundError:
            return (404, {}, "{}")

        return (200, {}, json.dumps(_json))

    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/devices/(\\d+)$"),
        callback=device_rules_callback,
    )


@pytest.fixture
def device_policies_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/20/policies",
        json=json.load(open("tests/securetrack/json/policies/20.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/400000/policies",
        status=404,
    )


@pytest.fixture
def devices_for_rule_test_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices",
        json=json.load(open("tests/securetrack/json/devices/for-rule-test.json")),
    )


@pytest.fixture
def search_rules_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search?devices=105&search_text=",
        json=json.load(open("tests/securetrack/json/rules/rule_search-105.json")),
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search/105?start=0&count=3000&search_text=",
        json=json.load(open("tests/securetrack/json/rules/rule_search-105-all-1.json")),
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search/105?start=3000&count=3000&search_text=",
        json=json.load(open("tests/securetrack/json/rules/rule_search-105-all-2.json")),
        match_querystring=True,
    )


@pytest.fixture
def rules_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/8/rules",
        json=json.load(open("tests/securetrack/json/rules/device-8.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/20/rules?add=documentation&uid=%7b3A1BA062-6B19-4C97-8F18-79CBA9EF0AA6%7d",
        json=json.load(open("tests/securetrack/json/rules/device-20-with-uid.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2285/rules?add=documentation&uid=Datacenter_access_in_@_8",
        json=json.load(
            open("tests/securetrack/json/rules/revision-2285-with-uid.json")
        ),
    )

    for i in (1, 5, 7, 20, 21):
        responses.add(
            responses.GET,
            f"https://198.18.0.1/securetrack/api/devices/{i}/rules?add=documentation",
            json=json.load(
                open(f"tests/securetrack/json/rules/device-{i}-add-documentation.json")
            ),
        )
    responses.add(
        responses.GET, "https://198.18.0.1/securetrack/api/devices/10/rules", status=500
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2285/rules",
        json=json.load(open("tests/securetrack/json/rules/device-8.json")),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2226/rules",
        status=500,
    )


@pytest.fixture
def revisions_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/8/latest_revision",
        json=json.load(
            open("tests/securetrack/json/revisions/device-8-latest-revision.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/100000",
        status=404,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/100000/latest_revision",
        status=404,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/100000/revisions",
        status=404,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/8/revisions",
        json=json.load(open("tests/securetrack/json/revisions/device-8.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2226",
        json=json.load(open("tests/securetrack/json/revisions/revision-2226.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2285",
        json=json.load(open("tests/securetrack/json/revisions/revision-2285.json")),
    )

    responses.add(
        responses.GET, "https://198.18.0.1/securetrack/api/revisions/400000", status=404
    )


@pytest.fixture
def network_objects_mock(devices_mock):
    def device_cb(request):
        device_id = request.path_url.split("/")[-2]
        params = parse_qs(urlparse(request.path_url).query)
        if params.get("add_parent_objects", None) == ["true"]:
            postfix = "-add-parent-objects"
        elif params.get("add_parent_objects", None) == ["false"]:
            postfix = "-no-parent-objects"
        else:
            postfix = ""

        fileString = f"tests/securetrack/json/network_objects/{device_id}{postfix}.json"

        if not os.path.exists(fileString):
            return (404, {}, "")
        else:
            return (200, {}, open(fileString).read())

    def search_cb(request):
        # key: value dictionary of params, but the value is a list in case there are multiple values for a single query param
        params = parse_qs(urlparse(request.path_url).query)
        if params.get("filter", [None])[0] == "uid":
            uid = params.get("uid", [None])[0]
            _json = Path(f"tests/securetrack/json/network_objects/{uid}.json")
            if _json.is_file():
                return (200, {}, _json.open().read())
        return (200, {}, json.dumps({"network_objects": {}}))

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/10/network_objects/80681",
        json=json.load(
            open(
                "tests/securetrack/json/network_objects/device_10_network_objects_80681.json"
            )
        ),
        status=200,
    )

    responses.add_callback(
        responses.GET,
        re.compile(r"https://198.18.0.1/securetrack/api/devices/\d+/network_objects"),
        callback=device_cb,
    )

    responses.add_callback(
        responses.GET,
        re.compile(r"https://198.18.0.1/securetrack/api/network_objects/search"),
        callback=search_cb,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/24522/network_objects?show_members=true&start=0&count=2000&get_total=true",
        json=json.load(
            open(
                "tests/securetrack/json/network_objects/revision_24522_network_objects.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/24522/network_objects/450118?show_members=true",
        json=json.load(
            open(
                "tests/securetrack/json/network_objects/revision_24522_network_objects_450118.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/network_objects/15/groups",
        json=json.load(
            open(
                "tests/securetrack/json/network_objects/network_objects_15_groups.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/network_objects/12/rules",
        json=json.load(
            open("tests/securetrack/json/network_objects/network_objects_12_rules.json")
        ),
    )


@pytest.fixture
def services_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/1/services",
        json=json.load(open("tests/securetrack/json/services/1.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/services/search",
        match=[
            matchers.query_string_matcher(
                "filter=uid&uid=3fbd8116-3801-4bee-8593-3cbf999da671"
            ),
        ],
        json=json.load(
            open(
                "tests/securetrack/json/services/search-3fbd8116-3801-4bee-8593-3cbf999da671.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/services/search",
        match=[
            matchers.query_string_matcher(
                "uid=3fbd8116-3801-4bee-8593-3cbf999da671&filter=uid&start=0&count=2000&get_total=true"
            )
        ],
        json=json.load(
            open(
                "tests/securetrack/json/services/search-3fbd8116-3801-4bee-8593-3cbf999da671.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/services/search?filter=uid&uid=3fbd8116-3801-4bee-8593-3cbf999da671&device_id=1",
        match_querystring=True,
        json=json.load(
            open(
                "tests/securetrack/json/services/search-device-1-3fbd8116-3801-4bee-8593-3cbf999da671.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/services/2961503/groups",
        json=json.load(
            open("tests/securetrack/json/services/services_2961503_groups.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/services/2937342/rules",
        json=json.load(
            open("tests/securetrack/json/services/services_2937342_rules.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/8/services?name=daytime&show_members=False&add_parent_objects=True&start=0&count=2000&get_total=true",
        json=json.load(
            open("tests/securetrack/json/services/devices_8_services_daytime.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/24522/services?name=MSExchange&show_members=False&add_parent_objects=True&start=0&count=2000&get_total=true",
        json=json.load(
            open(
                "tests/securetrack/json/services/revisions_24522_services_MSExchange.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/24522/services/3330108?show_members=false",
        json=json.load(
            open(
                "tests/securetrack/json/services/revisions_24522_services_3330108.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/8/services/3099979?show_members=false",
        json=json.load(
            open("tests/securetrack/json/services/devices_8_services_3099979.json")
        ),
    )


@pytest.fixture
def zone_subnets_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones/78/entries",
        json=json.load(open("tests/securetrack/json/zones/zone-entries-78.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones/200/entries",
        status=404,
    )


@pytest.fixture
def zone_descendants_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones/80/descendants",
        json=json.load(open("tests/securetrack/json/zones/zone-descendants-80.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones/200/descendants",
        status=404,
    )


@pytest.fixture
def zones_mock(st):
    def zone_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(open(f"tests/securetrack/json/zones/zone-{_id}.json"))
        except FileNotFoundError:
            return (404, {}, "{}")

        return (200, {}, json.dumps(_json))

    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/zones/(\\d+)$"),
        callback=zone_callback,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones",
        json=json.load(open("tests/securetrack/json/zones/zones.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/zones/internet_representing_address",
        body="8.8.4.4",
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/zones/internet_representing_address",
        status=200,
    )


@pytest.fixture()
def test_post_add_domain_json():
    json.load(open("tests/securetrack/json/domain/post_add_domain.json"))


@pytest.fixture
def domains_mock(st):
    def domain_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(open(f"tests/securetrack/json/domains/domain-{_id}.json"))
        except FileNotFoundError:
            return (404, {}, "{}")

        return (200, {}, json.dumps(_json))

    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/domains/(\\d+)$"),
        callback=domain_callback,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/domains",
        json=json.load(open("tests/securetrack/json/domains/domains.json")),
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/domains/",
        headers={"Location": "https://198.18.0.1/securetrack/api/domains/7"},
    )
    responses.add(responses.PUT, "https://198.18.0.1/securetrack/api/domains/7")


@pytest.fixture
def generic_interfaces_mock(st):
    def interface_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_interfaces/int-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Generic Interface Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    def management_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_interfaces/mgmt-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Management Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Interface by ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interface/(\\d+)$"
        ),
        callback=interface_callback,
    )

    """Get Interfaces by Management ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interface/mgmt/(\\d+)$"
        ),
        callback=management_callback,
    )

    """Delete Interface by ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interface/(\\d+)$"
        ),
        callback=interface_callback,
    )

    """Delete Interfaces by Management ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interface/mgmt/(\\d+)$"
        ),
        callback=management_callback,
    )


@pytest.fixture
def generic_route_mock(st):
    def route_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_routes/route-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Generic Route Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    def management_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_routes/mgmt-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Management Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Route by ID"""
    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/topology/generic/route/(\\d+)$"),
        callback=route_callback,
    )

    """Get Routes by Management ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/route/mgmt/(\\d+)$"
        ),
        callback=management_callback,
    )

    """Delete Route by ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile("https://198.18.0.1/securetrack/api/topology/generic/route/(\\d+)$"),
        callback=route_callback,
    )

    """Delete Routes by Management ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/route/mgmt/(\\d+)$"
        ),
        callback=management_callback,
    )


@pytest.fixture
def generic_vpn_mock(st):
    def vpn_callback(request):
        parsed_url = urlparse(request.url)
        parts = parsed_url.path.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_vpns/vpn-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Generic Vpn Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    def device_callback(request):
        parsed_url = urlparse(request.url)
        parts = parsed_url.path.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/generic_vpns/device-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": "Device Not Found"}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Vpn by ID"""
    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/topology/generic/vpn/(\\d+)$"),
        callback=vpn_callback,
    )

    """Get Vpns by Device ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            r"https://198.18.0.1/securetrack/api/topology/generic/vpn/device/(\d+)\?generic=true$"
        ),
        callback=device_callback,
    )

    """Get Vpns by Device ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            r"https://198.18.0.1/securetrack/api/topology/generic/vpn/device/(\d+)\?generic=false$"
        ),
        callback=device_callback,
    )

    """Delete Vpn by ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile("https://198.18.0.1/securetrack/api/topology/generic/vpn/(\\d+)$"),
        callback=vpn_callback,
    )

    """Delete Vpns by Device ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/vpn/device/(\\d+)$"
        ),
        callback=device_callback,
    )


@pytest.fixture
def generic_transparent_firewall_mock(st):
    def data_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(
                    f"tests/securetrack/json/generic_transparent_firewalls/data-{_id}.json"
                )
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": f"Layer2Data Id {_id} not found"}}),
            )

        return (200, {}, json.dumps(_json))

    def device_callback(request):
        parsed_url = urlparse(request.url)
        parts = parsed_url.path.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(
                    f"tests/securetrack/json/generic_transparent_firewalls/device-{_id}.json"
                )
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": f"DeviceId {_id} not found."}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Transparent Firewalls by Device ID"""
    responses.add_callback(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/generic/transparentfw/device/9?generic=false",
        callback=device_callback,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/topology/generic/transparentfw/device/404?generic=false",
        status=404,
    )

    """Delete Layer 2 Data by ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw/(\\d+)$"
        ),
        callback=data_callback,
    )

    """Delete Transparent Firewalls by Device ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/transparentfw/device/(\\d+)$"
        ),
        callback=device_callback,
    )


@pytest.fixture
def generic_ignored_interface_mock(st):
    def mgmt_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(
                    f"tests/securetrack/json/generic_ignored_interfaces/mgmt-{_id}.json"
                )
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": f"Management Id {_id} not found."}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Ignored Interfaces by Management ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface/mgmt/(\\d+)$"
        ),
        callback=mgmt_callback,
    )

    """Delete Ignored Interfaces by Management ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface/mgmt/(\\d+)$"
        ),
        callback=mgmt_callback,
    )


@pytest.fixture
def generic_interface_customer_tag_mock(st):
    def interface_customer_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(
                    f"tests/securetrack/json/generic_interface_customers/int-cust-{_id}.json"
                )
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps(
                    {"result": {"message": f"Interface Customer Tag {_id} not found."}}
                ),
            )

        return (200, {}, json.dumps(_json))

    def device_callback(request):
        parsed_url = urlparse(request.url)
        path_parts = parsed_url.path.split("/")
        _id = int(path_parts[-1])

        try:
            _json = json.load(
                open(
                    f"tests/securetrack/json/generic_interface_customers/device-{_id}.json"
                )
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": f"Device Id {_id} not found."}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Interface Customer Tag by ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer/(\\d+)$"
        ),
        callback=interface_customer_callback,
    )

    """Get Interface Customer Tags by Device ID"""
    responses.add_callback(
        responses.GET,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer/device/(\\d+)\?generic=(true|false)$"
        ),
        callback=device_callback,
    )

    """Delete Interface Customer Tag by ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer/(\\d+)$"
        ),
        callback=interface_customer_callback,
    )

    """Delete Interface Customer Tags by Device ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile(
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer/device/(\\d+)$"
        ),
        callback=device_callback,
    )


@pytest.fixture
def join_cloud_mock(st):
    def cloud_callback(request):
        parts = request.url.split("/")
        _id = int(parts[-1])

        try:
            _json = json.load(
                open(f"tests/securetrack/json/join_clouds/cloud-{_id}.json")
            )
        except FileNotFoundError:
            return (
                404,
                {},
                json.dumps({"result": {"message": f"Cloud Id {_id} not found."}}),
            )

        return (200, {}, json.dumps(_json))

    """Get Join Cloud by Cloud ID"""
    responses.add_callback(
        responses.GET,
        re.compile("https://198.18.0.1/securetrack/api/topology/join/clouds/(\\d+)$"),
        callback=cloud_callback,
    )

    """Delete Join Cloud by Cloud ID"""
    responses.add_callback(
        responses.DELETE,
        re.compile("https://198.18.0.1/securetrack/api/topology/join/clouds/(\\d+)$"),
        callback=cloud_callback,
    )


@pytest.fixture()
def change_windows_mock_fails():
    responses.add(
        responses.GET, "https://198.18.0.1/securetrack/api/change_windows", status=403
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks",
        status=403,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks/197",
        status=403,
    )


@pytest.fixture()
def change_windows_mock_fails_json():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows",
        body="No no no",
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks",
        body="No no no",
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks/197",
        body="No no no",
    )


@pytest.fixture()
def change_windows_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows?start=0&count=2000&get_total=true",
        json=json.load(
            open("tests/securetrack/json/change_windows/change_windows.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks",
        json=json.load(
            open("tests/securetrack/json/change_windows/change_window_tasks.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/change_windows/07c230ce-2dec-4109-a0db-33ff45ba1057/tasks/197",
        json=json.load(
            open("tests/securetrack/json/change_windows/change_window_task_197.json")
        ),
    )


@pytest.fixture
def applications_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/174/applications?start=0&count=2000&get_total=true",
        json=json.load(open("tests/securetrack/json/applications/174.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/206/applications?start=0&count=2000&get_total=true",
        json={"applications": {"count": 0, "application": []}},
    )


@pytest.fixture
def applications_failed_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/1337/applications?start=0&count=2000&get_total=true",
        body="No no no",
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/1338/applications?start=0&count=2000&get_total=true",
        status=500,
    )


@pytest.fixture
def rule_documentation_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/24522/rules/53419/documentation",
        json=json.load(
            open(
                "tests/securetrack/json/rules/revisions-24522-rules-53419-documentation.json"
            )
        ),
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securetrack/api/revisions/24522/rules/53419/documentation",
        status=200,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/revisions/24522/rules/53419/documentation",
        status=204,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/devices/264/rules/53419/documentation",
        status=204,
    )


@pytest.fixture
def rule_last_usage_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_last_usage/find_all/1",
        json=json.load(open("tests/securetrack/json/rules/rule_last_usage-1.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_last_usage/find/1/d5adc685-6498-47d0-ad62-f7eeae18a069",
        json=json.load(open("tests/securetrack/json/rules/rule_last_usage-1-1.json")),
    )


@pytest.fixture
def properties_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/properties",
        json=json.load(open("tests/securetrack/json/properties/properties.json")),
    )

    responses.add(
        responses.PUT,
        "https://198.18.0.1/securetrack/api/properties",
        status=200,
    )


@pytest.fixture
def time_object_mock():
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/1/time_objects",
        json=json.load(open("tests/securetrack/json/time_objects/1.json")),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2812/time_objects",
        json=json.load(open("tests/securetrack/json/time_objects/revision-2812.json")),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2812/time_objects/388",
        json=json.load(
            open("tests/securetrack/json/time_objects/time-objects-388.json")
        ),
    )
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/revisions/2812/time_objects/388,390",
        json=json.load(
            open("tests/securetrack/json/time_objects/time-objects-388-390.json")
        ),
    )


@pytest.fixture
def licenses_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/licenses",
        json=json.load(open("tests/securetrack/json/licenses/licenses.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/licenses/evaluation",
        json=json.load(open("tests/securetrack/json/licenses/license_evaluation.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/licenses/1",
        json=json.load(open("tests/securetrack/json/licenses/license_1.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/licenses/tiered-license",
        json=json.load(open("tests/securetrack/json/licenses/tiered_license.json")),
    )


@pytest.fixture
def security_policy_mock(st):
    def read_plain_text(file_path):
        with open(file_path, mode="r", encoding="utf-8") as f:
            return f.read()

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies?ignoreSecureTrack2Data=true",
        json=json.load(
            open(
                "tests/securetrack/json/security_policies/security_policies_classic.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies",
        json=json.load(
            open("tests/securetrack/json/security_policies/security_policies.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies/global",
        json=json.load(
            open(
                "tests/securetrack/json/security_policies/security_policies_global.json"
            )
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies/7131354762492230143/export",
        body=read_plain_text(
            "tests/securetrack/json/security_policies/security_policy_export_csv.txt"
        ),
        content_type="text/plain",
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies/404/export",
        status=404,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/security_policies/5/mapping",
        json=json.load(
            open(
                "tests/securetrack/json/security_policies/usp_device_interface_map.json"
            )
        ),
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/security_policies/5/manual_mapping",
        status=200,
    )

    responses.add(
        responses.POST,
        "https://198.18.0.1/securetrack/api/security_policies/404/manual_mapping",
        status=404,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/security_policies/7131354762492230143",
        status=204,
    )

    responses.add(
        responses.DELETE,
        "https://198.18.0.1/securetrack/api/security_policies/404",
        status=404,
    )


@pytest.fixture
def search_rules_on_open_tickets_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices",
        json=json.load(open("tests/securetrack/json/devices/rule_mod_devices.json")),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search?devices=1&search_text=inprogressticketid:*",
        json=json.load(
            open("tests/securetrack/json/rules/search_rules_in_open_tickets.json")
        ),
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search/?search_text=inprogressticketid:*",
        json=json.load(
            open("tests/securetrack/json/rules/search_rules_in_open_tickets.json")
        ),
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search/1?start=0&count=3000&search_text=inprogressticketid%3A%2A",
        json=json.load(
            open(
                "tests/securetrack/json/rules/search_rules_in_open_tickets_device_1.json"
            )
        ),
        match_querystring=True,
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/rule_search/1?start=3000&count=3000&search_text=inprogressticketid%3A%2A",
        json=json.load(
            open(
                "tests/securetrack/json/rules/search_rules_in_open_tickets_device_1.json"
            )
        ),
        match_querystring=True,
    )


@pytest.fixture
def violating_rules_mock(st):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/violating_rules/264/count",
        json=json.loads('{"count": {"count": 17}}'),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/violating_rules/264/device_violations?ignoreSecureTrack2Data=False&severity=CRITICAL&type=SECURITY_POLICY",
        json=json.load(
            open("tests/securetrack/json/rules/device_264_violating_rules.json")
        ),
    )

    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/violating_rules/1/device_violations?ignoreSecureTrack2Data=False&severity=LOW&type=SECURITY_POLICY",
        json=json.load(
            open("tests/securetrack/json/rules/device_1_violating_rules.json")
        ),
    )



from enum import Enum
from typing import List, Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop


@propify
class GenericTransparentFirewall(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int))
    output_l3_device_id: Optional[int] = prop(
        None, converter=optional(int), key="outputL3DeviceId"
    )
    output_l3_is_generic_device: Optional[bool] = prop(
        None, key="outputL3IsGenericDevice"
    )
    output_l3_interface_name: Optional[str] = prop(None, key="outputL3InterfaceName")
    layer2_device_id: Optional[int] = prop(
        None, converter=optional(int), key="layer2DeviceId"
    )
    input_l2_interface_name: Optional[str] = prop(None, key="inputL2InterfaceName")
    output_l2_interface_name: Optional[str] = prop(None, key="outputL2InterfaceName")
    input_l3_device_id: Optional[int] = prop(
        None, converter=optional(int), key="inputL3DeviceId"
    )
    input_l3_is_generic_device: Optional[bool] = prop(
        None, key="inputL3IsGenericDevice"
    )
    input_l3_interface_name: Optional[str] = prop(None, key="inputL3InterfaceName")



from enum import Enum
from typing import Optional

from pytos2.api import BaseAPI, get_app_api_session, boolify
from pytos2.utils import setup_logger

from requests import Response


LOGGER = setup_logger("scw_api")


class ScwAPI(BaseAPI):
    class Meta(Enum):
        PATH = "securechangeworkflow/api/securechange"
        APP = "SCW"
        TOS2_ENV = "SC_SERVER_SERVICE"

    def __init__(
        self, hostname: Optional[str], username: Optional[str], password: Optional[str]
    ):
        self.hostname, self.username, self.password, self.session = get_app_api_session(
            app=self, hostname=hostname, username=username, password=password
        )

    def get_workflow(
        self, _id: Optional[int] = None, name: Optional[str] = None
    ) -> Response:
        if _id and name:
            raise TypeError(
                "Either '_id' or 'name' can be passed as arguments, but not both."
            )
        params = boolify({"id": _id, "name": name})

        r = self.session.get("workflows", params=params)
        return r

    def redo_step(
        self,
        ticket_id: int,
        step_id: int,
        task_id: int,
        to_step_id: int,
        comment: str = "",
    ) -> Response:
        if not comment:
            comment = "Reassigned by script"

        body = {"redo_step_comment": {"comment": comment}}
        r = self.session.put(
            f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/redo/{to_step_id}",
            json=body,
        )
        return r

    def reassign_task(
        self,
        ticket_id: int,
        step_id: int,
        task_id: int,
        assignee_id: int,
        comment: str = "",
    ) -> Response:
        if not comment:
            comment = "Reassigned by script"

        body = {"reassign_task_comment": {"comment": comment}}
        r = self.session.put(
            f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/reassign/{assignee_id}",
            json=body,
        )
        return r

    def put_task(self, ticket_id: int, task_id: int, task: dict) -> Response:
        """This function intentionally doesn't have a step_id argument, because you can only put tasks at the current step"""
        r = self.session.put(
            f"tickets/{ticket_id}/steps/current/tasks/{task_id}", json=task
        )
        return r

    def reject_ticket(
        self,
        ticket_id: int,
        handler_id: Optional[int] = None,
        comment: Optional[str] = "",
    ) -> Response:
        params = {} if handler_id is None else {"handler_id": handler_id}
        comment = "None provided" if not comment else comment
        body = {"reject_comment": {"comment": comment}}
        r = self.session.put(f"tickets/{ticket_id}/reject", params=params, json=body)
        return r

    def put_field(
        self, ticket_id: int, task_id: int, field_id: int, field: dict
    ) -> Response:
        """This function intentionally doesn't have a step_id argument, because you can only put tasks at the current step"""
        r = self.session.put(
            f"tickets/{ticket_id}/steps/current/tasks/{task_id}/fields/{field_id}",
            json=field,
        )
        return r

    def put_fields(self, ticket_id: int, task_id: int, fields: dict) -> Response:
        """This function intentionally doesn't have a step_id argument, because you can only put tasks at the current step"""
        r = self.session.put(
            f"tickets/{ticket_id}/steps/current/tasks/{task_id}/fields", json=fields
        )
        return r

    def cancel_ticket(
        self, ticket_id: int, requestor_id: Optional[int] = None
    ) -> Response:
        params = {} if requestor_id is None else {"requestor_id": requestor_id}
        r = self.session.put(f"tickets/{ticket_id}/cancel", params=params)
        return r

    def confirm_ticket(
        self,
        ticket_id: int,
        comment: Optional[str] = "",
        requestor_id: Optional[int] = None,
    ) -> Response:
        comment = "None provided" if not comment else comment
        params = {} if requestor_id is None else {"requestor_id": requestor_id}
        r = self.session.put(
            f"tickets/{ticket_id}/confirm",
            params=params,
            json={"confirm_comment": {"comment": comment}},
        )
        return r



from enum import Enum
from typing import Optional, List

from attr.converters import optional
import attr

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify, get_api_node

import netaddr  # type: ignore
from netaddr import IPAddress, IPNetwork, IPRange  # type: ignore

from pytos2 import securetrack


class InterfaceXsiType(Enum):
    INTERFACE = "interfaceDTO"


@propify
class InterfaceIP(Jsonable):
    class Prop(Enum):
        IP = "ip"
        NETMASK = "netmask"

    netmask: Optional[IPNetwork] = prop(None)
    ip: Optional[IPAddress] = prop(None)
    visibility: Optional[str] = prop(None)
    precedence: Optional[str] = prop(None)


@propify
class Interface(Jsonable):
    class Meta(Enum):
        ROOT = "interface"

    class Prop(Enum):
        GLOBAL = "global"

    id: int = prop(None)
    uid: Optional[str] = prop(None)
    device_id: Optional[int] = prop(None, converter=optional(int))
    name: str = prop(None)
    is_global: bool = prop(False, key=Prop.GLOBAL.value)
    zone_id: int = prop(None)
    implicit: bool = prop(False, repr=False)
    interface_ips: List[IPAddress] = prop(factory=list, repr=False)
    int_ip_list: List[IPAddress] = prop(factory=list)
    acl_name: str = prop(None)
    direction: str = prop(None)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        int_list = _obj.get("interface_ips")

        if len(int_list) > 0:
            try:
                raw_ips = int_list.get("interface_ip")
                for i in raw_ips:
                    ip_addr = f"{i['ip']}/{IPAddress(i['netmask']).netmask_bits()}"
                    kwargs["int_ip_list"].append(ip_addr)
            except Exception as e:  # pragma: no cover
                pass

        return cls(**kwargs)


class BindableObjectXsiType(Enum):
    BINDABLE_OBJECT = "BindableObjectDTO"


@propify
class BindableObject(Interface):
    class Prop(Enum):
        GLOBAL = "global"

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        int_list = _obj.get("interface_ips")

        if len(int_list) > 0:
            try:
                raw_ips = int_list.get("interface_ip")
                for i in raw_ips:
                    ip_addr = f"{i['ip']}/{IPAddress(i['netmask']).netmask_bits()}"
                    kwargs["int_ip_list"].append(ip_addr)
            except Exception as e:  # pragma: no cover
                pass

        return cls(**kwargs)


@propify
class TopologyInterface(Jsonable):
    class Meta(Enum):
        ROOT = "interface"

    class Prop(Enum):
        IP_ADDRESS = "ip"
        NETMASK = "mask"
        CIDR = "cidr"

    ip: Optional[IPAddress] = prop(None)
    device_id: Optional[int] = prop(None, converter=optional(int))
    virtual_router: Optional[str] = prop(None)
    generic_device_id: Optional[int] = prop(None)
    name: str = prop(None)
    zone: Optional[int] = prop(None)
    mask: Optional[IPAddress] = prop(None, repr=False)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        kwargs["ip"] = IPNetwork(f"{_obj['ip']}/{_obj.get('mask', '32')}")

        return cls(**kwargs)


@propify
class PolicyZone(Jsonable):
    is_global: Optional[bool] = prop(False)
    name: Optional[str] = prop(None)



import json
from datetime import datetime, time as datetime_time
from typing import List, Union
from unittest import mock

from netaddr import IPNetwork, IPAddress, IPRange  # type: ignore
import pytest  # type: ignore
import responses  # type: ignore


from pytos2.securechange.fields import (
    AtType,
    AccessRequest,
    CloneServerPolicyRequest,
    Field,
    IPObject,
    MultiAccessRequest,
    MultiServerDecommisionRequest,
    MultiGroupChange,
    Target,
    TextField,
    TextArea,
    MultiTextField,
    MultiTextArea,
    Manager,
    Hyperlink,
    MultiHyperlink,
    MultiTarget,
    MultiService,
    MultiNetworkObject,
    Date,
    Time,
    ApproveReject,
    Checkbox,
    DropDownList,
    MultipleSelection,
    Zone,
    classify_group_change_member_from_api,
    GroupChange,
    GroupChangeMember,
    GroupChangeMemberHost,
    GroupChangeMemberNetwork,
    GroupChangeMemberRange,
    RuleModificationField,
    RuleRecertification,
)
from pytos2.models import UnMapped
from pytos2.securechange.ticket import Ticket
from pytos2.securechange.designer import DesignerResults
from pytos2.securechange.service import PredefinedServiceName
from pytos2.securechange.fields import Service
from pytos2.securechange.fields.rule_operation import RuleKey
from pytos2.securechange.rule import SlimRule
from pytos2.securechange.network_object import HostNetworkObject
from pytos2.securechange.fields.rule_operation import (
    ModificationTransportService,
    ModificationServiceObject,
    ModificationIcmpService,
    ModificationIPService,
    RangeDTO,
)

from pytos2.securetrack.device import Device
from pytos2.securetrack.rule import SecurityRule, BindingPolicy
from pytos2.securetrack.service_object import (
    TCPServiceObject,
    ICMPServiceObject,
    Service as STService,
    OtherIPServiceObject,
)


@pytest.fixture
def mgc(get_test_field):
    return get_test_field(MultiGroupChange)


class TestField:
    @pytest.fixture
    def generic_field(self):
        return Field.kwargify({"name": "name", "id": 123})

    def test_json_override(self, generic_field):
        assert generic_field._json == generic_field.data
        generic_field._json = {}
        assert generic_field._json == {}

    def test_read_only(self):
        field = TextArea.kwargify(
            json.load(open("tests/securechange/json/field/read_only-field.json"))
        )
        assert len(field._json) == 0

    def test_dirty(self, generic_field):
        generic_field.id = 124
        assert generic_field._dirty


class TestMultiAccessRequest:
    @pytest.fixture
    def ar(self, get_test_field):
        return get_test_field(MultiAccessRequest)

    def test_attrs(self, ar):
        assert hasattr(ar, "designer_result")
        assert isinstance(ar.access_requests, list)

    @responses.activate
    def test_designer_results(self, ar, ar_with_designer_results):
        responses.add(
            responses.GET,
            "https://10.20.50.192/securechangeworkflow/api/securechange/tickets/251/steps/1242/tasks/1254/fields/21516/designer",
            json=json.load(
                open("tests/securechange/json/designer/designer_result_251.json")
            ),
        )

        assert isinstance(
            ar_with_designer_results.designer_result.get_results(), DesignerResults
        )

        assert ar.designer_result.get_results() is None

    def test_add_arr(self, ar):
        new_ar = ar.add_ar()
        assert new_ar.sources == []


class TestAccessRequest:
    @pytest.fixture
    def ar1(self, get_test_field):
        return get_test_field(MultiAccessRequest).access_requests[0]

    def test_add_source(self, ar1):
        src = ar1.add_source("1.2.3.4")
        assert src.ip_address == "1.2.3.4"
        assert ar1.comment == "test comment mapping"

    def test_add_destination(self, ar1):
        dst = ar1.add_destination("4.3.2.1")
        assert dst.ip_address == "4.3.2.1"

    def test_add_tcp_service(self, ar1):
        svc = ar1.add_service("tcp 443")
        assert svc.port == "443"
        assert svc.protocol == Service.Protocol.TCP

    def test_add_predefined_service(self, ar1):
        svc = ar1.add_service("https")
        assert svc.name == PredefinedServiceName.HTTPS

    def test_add_invalid_predefined_service(self, ar1):
        with pytest.raises(ValueError):
            ar1.add_service("httpoo")

    @responses.activate
    def test_add_object(self, ar1, network_objects_mock):
        dst = ar1.add_destination(name="192.168.1.82V", device=174)
        assert dst.type.value == "host"

    @responses.activate
    def test_add_target(self, st, ar1, mock_devices, devices_mock):
        policy_name = "Standard"

        def test_target(target: Target, target_name: str, type: Target.ObjectType):
            assert target.type.value == type.value
            assert target.name == target_name

        def test_policy(device: Device, target_name: str, type: Target.ObjectType):
            policy = st.get_device_policy(device=device.id, policy=policy_name)
            assert isinstance(policy, BindingPolicy)
            target = ar1.add_target(device=device, policy=policy)
            test_target(target, target_name, type)

        def add_target(
            device: Device,
            target_name: str,
            type: Target.ObjectType,
        ):
            t1 = ar1.add_target(device=device, policy=policy_name)
            t2 = ar1.add_target(device=device.id, policy=policy_name)
            t3 = ar1.add_target(device=device.name, policy=policy_name)
            test_target(t1, target_name, type)
            test_target(t2, target_name, type)
            test_target(t3, target_name, type)

            if device.model is not Device.Model.ASA:
                test_policy(device, target_name, type)

            ar1.targets = []

        def device_not_found():
            target = ar1.add_target(device="Nonexistent", policy="Test")
            assert target is None

        def expect_exception(device, error, policy=policy_name):
            with pytest.raises(ValueError) as exception:
                ar1.add_target(device=device, policy=policy)

            assert error in str(exception.value)

        def test_devices(devices):
            for d in devices:
                responses.add(
                    responses.GET,
                    f"https://198.18.0.1/securetrack/api/devices/{d.id}/policies",
                    json=json.load(open("tests/securetrack/json/policies/20.json")),
                )
                device = st.get_device(d.id)
                assert isinstance(device, Device)

                name = ""
                type = None
                if device.model is Device.Model.ASA:
                    ar1.use_topology = False
                    assert ar1.use_topology is False
                    error = f"ACL name must be specified in the policy argument to add a {device.model.value} target to an AR with topology disbaled"
                    expect_exception(device, error, "")

                    name = policy_name
                    type = Target.ObjectType.ACL
                    add_target(device, name, type)

                    ar1.use_topology = True
                    assert ar1.use_topology is True
                    name = device.name
                    type = Target.ObjectType.FIREWALL
                    add_target(device, name, type)

                elif device.model in (
                    Device.Model.PANORAMA_NG_FW,
                    Device.Model.FMG_FIREWALL,
                ):
                    ar1.use_topology = False
                    assert ar1.use_topology is False
                    error = f"Cannot add a {device.model.value} target to an AR with topology disabled"
                    expect_exception(device, error)

                    ar1.use_topology = True
                    assert ar1.use_topology is True
                    name = device.name
                    type = Target.ObjectType.FIREWALL
                    add_target(device, name, type)

                elif device.model in (Device.Model.PANORAMA_DEVICE_GROUP,):
                    ar1.use_topology = True
                    assert ar1.use_topology is True
                    error = f"Cannot add a {device.model.value} target to an AR with topology enabled"
                    expect_exception(device, error)

                    ar1.use_topology = False
                    assert ar1.use_topology is False

                    name = f"{None}>{None}"
                    type = Target.ObjectType.ZONE_TO_ZONE
                    add_target(device, name, type)

                elif device.model in (Device.Model.FMG_ADOM, Device.Model.FMG_VDOM):
                    ar1.use_topology = True
                    assert ar1.use_topology is True
                    error = f"Cannot add a {device.model.value} target to an AR with topology enabled"
                    expect_exception(device, error)

                    # test for pytos2/securechange/fields/__init__.py line 767
                    # when no policy is provided this is
                    # unreachable since value error will be thrown
                    # after not finding a policy with get_policy

                    ar1.use_topology = False
                    assert ar1.use_topology is False
                    # error = "policy argument must be specified to add this type of target to this AR"
                    # expect_exception(device, error, "")

                    name = f"{None}>{None}"
                    type = Target.ObjectType.ZONE_TO_ZONE
                    add_target(device, name, type)

                elif device.model in (Device.Model.MODULE, Device.Model.MODULE_CLUSTER):
                    name = device.name
                    type = Target.ObjectType.FIREWALL
                    add_target(device, name, type)

                elif device.model in (Device.Model.CP_CMA, Device.Model.CP_SMRT_CNTR):
                    name = policy_name
                    type = Target.ObjectType.POLICY
                    add_target(device, name, type)

                else:
                    error = f"Support for {device.model.value} type targets is not been implemented."
                    expect_exception(device, error)
                    continue

        device_not_found()
        test_devices(mock_devices)

        # Testing ASA
        device = st.get_device(8)

        ar1.use_topology = True
        o = ar1.add_target(device=device)
        assert o.name == "ASAv"
        assert o.type is Target.ObjectType.FIREWALL

        # Testing ASA with use_topology off
        ar1.use_topology = False
        o = ar1.add_target(device=device, policy="TESTME")
        assert o.name == "TESTME"
        assert o.management_name == "ASAv"
        assert o.type is Target.ObjectType.ACL

    def test_jsonify(self):
        ar = AccessRequest()
        assert "sources" not in ar._json

    @responses.activate
    def test_target_count(self, st, ar1, mock_devices, devices_mock):
        policy_name = "Standard"

        def check_target_count(targets: List[Target]):
            assert len(targets) == len(ar1.targets)

        def add_targets(devices):
            targets = []
            ar1.targets = []
            for d in devices:
                responses.add(
                    responses.GET,
                    f"https://198.18.0.1/securetrack/api/devices/{d.id}/policies",
                    json=json.load(open("tests/securetrack/json/policies/157.json")),
                )
                device = st.get_device(d.id)
                assert isinstance(device, Device)

                if device.model in (
                    Device.Model.PANORAMA_NG_FW,
                    Device.Model.FMG_FIREWALL,
                ):
                    ar1.use_topology = True
                else:
                    ar1.use_topology = False

                target = ar1.add_target(
                    device=device,
                    policy="FG_SITE-A-POLICY",
                    source_zone="any",
                    destination_zone="port2",
                )
                target = ar1.add_target(
                    device=device,
                    policy="FG_SITE-B-POLICY",
                    source_zone="any",
                    destination_zone="port2",
                )
                targets.append(target)

            check_target_count(targets)

        pan = [
            d
            for d in mock_devices
            if (
                d.id in (174, 179, 180, 181, 182)
                and d.model is not Device.Model.PANORAMA_NG
            )
        ]
        add_targets(pan)

        fm = [d for d in mock_devices if d.id in (157, 159, 160, 161)]
        add_targets(fm)


class TestCloneServerPolicyRequest:
    @pytest.fixture
    def server_policy_request(self, get_test_field):
        return get_test_field(CloneServerPolicyRequest)

    def test_attrs(self, server_policy_request):
        assert isinstance(server_policy_request.to_servers, list)
        assert isinstance(server_policy_request.to_servers[0], IPObject)
        assert server_policy_request.from_server.ip_address == "192.168.2.200"


class TestMultiServerDecommisionRequest:
    @pytest.fixture
    def server_decom(self, get_test_field):
        return get_test_field(MultiServerDecommisionRequest)

    def test_attrs(self, server_decom):
        assert isinstance(server_decom.server_decommission_requests, list)
        assert (
            server_decom.server_decommission_requests[0].comment
            == "Example Server Decom"
        )

    def test_add_server(self, server_decom):
        server_decom.server_decommission_requests[0].add_server("1.2.3.4")
        assert (
            server_decom.server_decommission_requests[0].servers[-1].ip_address
            == "1.2.3.4"
        )

    def test_add_bad_server(self, server_decom):
        with pytest.raises(ValueError):
            server_decom.server_decommission_requests[0].add_server("foobar")

    def test_kwargify_ticket(self, ticket_488):
        ticket = Ticket.kwargify(ticket_488)
        assert ticket.id == 488


class TestMultiGroupChange:
    def test_attrs(self, mgc):
        assert hasattr(mgc, "group_changes")
        assert isinstance(mgc.group_changes, list)

    def test_delete_member(self, mgc):
        member = mgc.group_changes[7].members[0]
        with pytest.raises(AssertionError):
            member.delete()
        member = mgc.group_changes[2].members[0]
        deleted_member = member.delete()
        assert isinstance(deleted_member, GroupChangeMemberHost)
        assert deleted_member.id == 1130
        assert deleted_member.management_id == 5
        assert str(deleted_member.ip) == "10.10.20.50"
        assert deleted_member.status.value == "DELETED"
        re_deleted_member = member.delete()
        assert re_deleted_member == deleted_member


class TestGroupChange:
    def test_address_book(self, mgc):
        assert str(mgc.group_changes[3].address_book) == "global"

    def test_zone(self, mgc):
        assert str(mgc.group_changes[7].zone) == "ssl.External"


class TestAddGroupChange:
    def test_bad_args(self, mgc):
        with pytest.raises(ValueError):
            mgc.add_group_change("thing")

    @responses.activate
    def test_bad_device(self, network_objects_mock, mgc):
        with pytest.raises(ValueError):
            mgc.add_group_change("bad", 1000000)

    @responses.activate
    def test_new_group(self, network_objects_mock, mgc: MultiGroupChange):
        res = mgc.add_group_change("new_group", 157)
        assert isinstance(res, GroupChange)

    @responses.activate
    def test_bad_uid(self, network_objects_mock, mgc):
        with pytest.raises(ValueError):
            mgc.add_group_change(uid="afb2d78d-356a-4a86-805a-000000000000")

    @responses.activate
    def test_bad_object_type(self, network_objects_mock, mgc):
        with pytest.raises(ValueError):
            mgc.add_group_change(uid="afb2d78d-356a-4a86-805a-c80afee8e338")

    @responses.activate
    def test_existing_group(self, network_objects_mock, mgc):
        res = mgc.add_group_change("New_group", 243)
        assert isinstance(res, GroupChange)


class TestAddMember:
    @pytest.fixture
    def gc(self, get_test_field):
        return get_test_field(MultiGroupChange).group_changes[9]

    @pytest.fixture
    def gc1(self, get_test_field):
        return get_test_field(MultiGroupChange).group_changes[0]

    @pytest.fixture
    def gc2(self, get_test_field):
        return get_test_field(MultiGroupChange).group_changes[2]

    @pytest.mark.parametrize(
        "name, details_arg, details, at_type, type",
        [
            (
                "newhost",
                "1.2.3.4",
                "1.2.3.4",
                GroupChangeMemberHost,
                GroupChangeMember.ObjectType.HOST,
            ),
            (
                "newnetwork",
                "1.2.3.0/24",
                "1.2.3.0/255.255.255.0",
                GroupChangeMemberNetwork,
                GroupChangeMember.ObjectType.NETWORK,
            ),
            (
                "newrange",
                "2.2.2.2-3.3.3.3",
                "[ 2.2.2.2 - 3.3.3.3 ]",
                GroupChangeMemberRange,
                GroupChangeMember.ObjectType.ADDRESS_RANGE,
            ),
        ],
    )
    def test_new(self, gc, name, details_arg, details, at_type, type):
        member = gc.add_member(name=name, details=details_arg)
        assert isinstance(member, GroupChangeMember)
        assert member.details == details
        assert isinstance(member, at_type)
        assert member.type is type

    def test_existing(self, gc):
        add_res = gc.add_member(name="new_net", details="1.3.4.0/255.255.255.0")
        assert add_res is None

    @responses.activate
    def test_uid(self, gc, network_objects_mock):
        add_res = gc.add_member(uid="20dcde39-b1ac-4013-adbb-81a5a705801d")
        assert isinstance(add_res, GroupChangeMember)
        assert add_res.uid == "20dcde39-b1ac-4013-adbb-81a5a705801d"

    @responses.activate
    def test_already_added_uid(self, gc2, network_objects_mock):
        add_res = gc2.add_member(uid="b4e77d4c-f471-4bb2-bfff-69503bdd6669")
        assert add_res is None

    @responses.activate
    def test_deleted_uid(self, gc1, network_objects_mock):
        add_res = gc1.add_member(uid="b17b4120-07d4-0875-d4b2-06227767ff57")
        assert isinstance(add_res, GroupChangeMember)

    @responses.activate
    def test_uid_missing(self, gc1, network_objects_mock):
        with pytest.raises(ValueError):
            gc1.add_member(uid="b17b4120-07d4-0875-d4b2-000000000000")

    @responses.activate
    def add_target_missing(self, mgc, network_objects_mock):
        group_change = GroupChange(name="moo", management_id=100000000)
        with pytest.raises(ValueError):
            group_change.add_member(uid="1111111111111111111111111111")

    def test_new_from_other_group(self, gc):
        add_res = gc.add_member(name="nost", details="3.3.3.3")
        assert isinstance(add_res, GroupChangeMember)
        assert add_res.at_type is AtType.OBJECT
        add_res = gc.add_member(name="nost", details="3.3.3.3")
        assert add_res is None

    def test_existing_mismatch(self, gc):
        with pytest.raises(ValueError):
            gc.add_member(name="nost", details="2.2.2.2")
        with pytest.raises(ValueError):
            gc.add_member(name="new_net", details="1.3.5.0/24")

    def test_bad_member(self, gc):
        with pytest.raises(ValueError):
            gc.add_member(name="bad", details="300.10.10.0")

    def test_args(self, gc):
        with pytest.raises(TypeError):
            gc.add_member(name="thing")


class TestGroupChangeMember:
    @pytest.fixture
    def mgc(self, get_test_field):
        return get_test_field(MultiGroupChange)

    def test_host(self, mgc):
        member = mgc.group_changes[0].members[0]
        assert member.details == "1.1.1.1"
        assert member.ip == IPAddress("1.1.1.1")
        member.ip = IPAddress("1.2.3.5")
        assert member.details == "1.2.3.5"
        with pytest.raises(ValueError):
            member.ip = "garbage"

    def test_network(self, mgc):
        member = mgc.group_changes[5].members[0]
        assert member.details == "10.10.103.0/255.255.255.0"
        assert member.network == IPNetwork("10.10.103.0/24")
        member.network = IPNetwork("10.12.14.0/24")
        assert member.details == "10.12.14.0/255.255.255.0"
        with pytest.raises(ValueError):
            member.network = "garbage"

    def test_range(self, mgc):
        member = mgc.group_changes[6].members[3]
        assert member.details == "[ 127.0.0.1 - 127.255.255.255 ]"
        assert member.range == IPRange("127.0.0.1", "127.255.255.255")
        member.range = "6.6.6.20-6.6.6.40"
        assert member.details == "[ 6.6.6.20 - 6.6.6.40 ]"
        with pytest.raises(ValueError):
            member.range = "garbage"

    def test_classify(self):
        assert isinstance(
            classify_group_change_member_from_api({"some": "field"}), UnMapped
        )


def test_approve_reject(get_test_field):
    field = get_test_field(ApproveReject)
    assert field.approved is bool(field) is True
    field.reject()
    assert field.approved is False
    assert field.reason == "None provided"
    field.approve("because")
    assert field.approved is True
    assert field.reason == "because"


def test_check_box(get_test_field):
    field = get_test_field(Checkbox)
    assert field.value is field.checked is bool(field) is True
    field.uncheck()
    assert field.value is False
    field.check()
    assert field.value is True
    field.toggle()
    assert field.value is False


class TestDate:
    @pytest.fixture
    def date(self, get_test_field):
        return get_test_field(Date)

    def test_value(self, date):
        assert date.value == "2019-05-08"

    def test_set_date(self, date):
        date.date = datetime(1988, 9, 11)
        assert date.value == "1988-09-11"

    def test_set_str_date(self, date):
        date.date = "1988-09-11"
        assert date.date == datetime(1988, 9, 11).date()

    def test_invalid_set_date(self, date):
        with pytest.raises(TypeError):
            date.date = "198-9-11"


class TestTime:
    @pytest.fixture
    def time(self, get_test_field):
        return get_test_field(Time)

    def test_value(self, time):
        assert time.value == "12:00"

    def test_set_time(self, time):
        time.time = datetime_time(5, 30)
        assert time.value == "05:30"

    def test_invalid_set_time(self, time):
        with pytest.raises(TypeError):
            time.time = "05:30"


class TestRuleRecertification:
    @responses.activate
    def test_is_invalid_rule(
        self, device_20_rules, device_1_rules, search_rules_on_open_tickets_mock
    ):
        recert = RuleRecertification()

        rules = [device_20_rules[0], device_20_rules[1], device_1_rules[0]]

        for rule in rules:
            recert.add_rule(rule)

        for rule in rules:
            is_invalid = recert.is_invalid_rule(rule)
            assert is_invalid is False

    def test_add_rule(self, device_20_rules, device_1_rules):
        recert = RuleRecertification()

        rule = device_20_rules[0]
        rule2 = device_20_rules[1]
        rule3 = device_1_rules[0]

        recert.add_rule(rule)
        recert.add_rule(rule2)
        recert.add_rule(rule3)

        rules = recert.get_rules()
        assert len(rules) == 3

    def test_add_rules(self, device_20_rules, device_1_rules):
        recert = RuleRecertification()

        rules = [device_20_rules[0], device_20_rules[1], device_1_rules[0]]

        recert.add_rules(rules)

        rules = recert.get_rules()
        assert len(rules) == 3


class TestRuleModification:
    @pytest.fixture
    def rule_mod(self, get_test_field):
        return get_test_field(RuleModificationField)

    def test_id(self, rule_mod):
        assert rule_mod.id == 24670
        assert rule_mod.rule_modifications[0].id == 3
        assert rule_mod.rule_modifications[0].source_modifications.id == 4

    def test_rule_key(self, rule_mod):
        rule_modification = rule_mod.rule_modifications[0]
        assert isinstance(rule_modification.rule_key, RuleKey)
        assert rule_modification.rule_key == RuleKey(
            device_id=20,
            binding_uid="{3A1BA062-6B19-4C97-8F18-79CBA9EF0AA6}",
            rule_uid="{C610F0C8-C339-11E4-9E5F-7F0000011414}",
        )

    def test_get_rules(self, rule_mod):
        rules = rule_mod.get_rules()
        assert len(rules) == 1
        assert isinstance(rules[0], SlimRule)

    def test_is_rule_added(self, rule_mod):
        rule = SecurityRule(uid="{C610F0C8-C339-11E4-9E5F-7F0000011414}")
        rule2 = SecurityRule(uid="{C610F0C8-C339-11E4-9E5F-7F0000011434}")

        assert rule_mod.is_rule_added(rule)
        assert not rule_mod.is_rule_added(rule2)

    def test_add_rule(self, device_20_rules, device_20_network_objects):
        rule_mod = RuleModificationField()

        rule = device_20_rules[0]

        rule_mod.add_destination_object(
            rule, HostNetworkObject(id=0, name="TestObject", ip="1.2.3.4")
        )

        rules = rule_mod.get_rules()
        assert len(rules) == 1

        rule_mod.add_source_object(device_20_rules[1], device_20_network_objects[0])

        rules = rule_mod.get_rules()
        assert len(rules) == 2

        rule_mod.add_service(
            device_20_rules[1],
            ModificationTransportService(
                name="Test", comment="Test2", port=RangeDTO(from_=443, to=448)
            ),
        )

    def test_removes(self, device_20_rules, device_20_network_objects):
        rule_mod = RuleModificationField()

        rule = device_20_rules[0]

        # 84 = SMCPM
        rule_mod.remove_destination_object(rule, device_20_network_objects[84])
        assert len(rule_mod.get_rules()) == 1

        # Some random network object that isn't included in the rule
        assert not rule_mod.remove_destination_object(
            device_20_rules[1], device_20_network_objects[50]
        )

        # 137 = m01
        rule_mod.remove_source_object(
            device_20_rules[1], device_20_network_objects[137]
        )
        assert len(rule_mod.get_rules()) == 2

    @responses.activate
    def test_rule_prop(self, device_20_rules, device_20_rules_mock):
        rule = device_20_rules[0]
        rule_mod = RuleModificationField()

        rule_mod.add_destination_object(
            rule, HostNetworkObject(id=0, name="TestObject", ip="1.2.3.4")
        )

        mod = rule_mod.rule_modifications[0]
        assert mod.rule.uid == rule.uid
        mod._rule = None
        assert mod.rule.uid == rule.uid

    def test_from_securetrack(self):
        obj = TCPServiceObject(
            comment="Hello Comment",
            name="Hello Name",
            protocol=6,
            min_port=443,
            max_port=443,
        )
        o = ModificationServiceObject.from_securetrack(obj)

        assert o.port.from_ == 443
        assert o.port.to == 443
        assert o.protocol == "TCP"
        assert isinstance(o, ModificationTransportService)

        obj = ICMPServiceObject(
            comment="icmp", name="icmp srvc", min_port=40, max_port=40
        )
        o = ModificationServiceObject.from_securetrack(obj)

        assert o.type.from_ == 40
        assert o.type.to == 40
        assert isinstance(o, ModificationIcmpService)

        obj = STService(comment="hello test", name="blah blah")
        o = ModificationServiceObject.from_securetrack(obj)
        assert isinstance(o, ModificationServiceObject)

        obj = OtherIPServiceObject(name="other ip server object")
        o = ModificationServiceObject.from_securetrack(obj)
        assert isinstance(o, ModificationIPService)


def test_drop_down_list(get_test_field):
    field = get_test_field(DropDownList)
    assert isinstance(field.options, list)


def test_hyperlink(get_test_field):
    field = get_test_field(Hyperlink)
    assert field.url == "http://"


def test_multi_hyperlink(get_test_field):
    field = get_test_field(MultiHyperlink)
    assert all(isinstance(h, Hyperlink) for h in field.hyperlinks)


def test_manager(get_test_field):
    field = get_test_field(Manager)
    assert isinstance(field.text, str)


def test_multiple_selection(get_test_field):
    field = get_test_field(MultipleSelection)
    # TODO test comparison
    assert isinstance(field.options, list) and isinstance(field.selected_options, list)


def test_multi_network_object(get_test_field):
    field = get_test_field(MultiNetworkObject)
    assert isinstance(field.network_objects, list)


def test_multi_service(get_test_field):
    field = get_test_field(MultiService)
    assert isinstance(field.services, list)


def test_multi_target(get_test_field):
    field = get_test_field(MultiTarget)
    assert isinstance(field.targets, list)


def test_text_field(get_test_field):
    field = get_test_field(TextField)
    assert field.text == "test"


def test_multi_text_field(get_test_field):
    field = get_test_field(MultiTextField)
    assert [t.text for t in field.text_fields] == ["a", "b"]


def test_text_area(get_test_field):
    field = get_test_field(TextArea)
    assert field.text == "test"


def test_multi_text_area(get_test_field):
    field = get_test_field(MultiTextArea)
    assert [t.text for t in field.text_areas] == ["1", "2"]



from datetime import datetime, date
from enum import Enum
from typing import Optional, List
from copy import deepcopy

from attr.converters import optional

from pytos2.models import Jsonable, IPType
from pytos2.utils import jsonify, propify, prop, safe_iso8601_date

from pytos2.securetrack.network_object import NetworkObject, Interface, ObjectReference

from .device import Device


class SecurityRuleType(Enum):
    CLOUD_SECURITY_RULE = "cloudSecurityRuleDTO"
    SECURITY_RULE = "securityRuleDTO"


@propify
class Track(Jsonable):
    class Level(Enum):
        DISABLED = "DISABLED"
        LOG = "LOG"
        NONE = "NONE"

    level: Optional[Level] = prop(None, repr=False)


@propify
class PolicyTarget(Jsonable):
    pass


@propify
class InstallableTarget(Jsonable):
    device_id: Optional[int] = prop(None, converter=optional(int))
    firewall_id: Optional[str] = prop(None)
    virtual_firewall_name: Optional[str] = prop(None, repr=False)
    virtual_firewall_securetrack_name: Optional[str] = prop(None)


@propify
class Zone(Jsonable):
    class Prop(Enum):
        GLOBAL = "global"

    name: Optional[str] = prop(None, repr=False)
    is_global: Optional[bool] = prop(None, repr=False, key=Prop.GLOBAL.value)


@propify
class PolicyXsiType(Enum):
    MANAGEMENT_POLICY = "managementPolicyDTO"


@propify
class BindingPolicy(Jsonable):
    class Prop(Enum):
        UNIQUE_ACTIVE_IN_ITG = "unique_active_in_itg"
        POLICY_TARGETS = "policyTargets"
        INSTALLABLE_TARGET = "installable_target"

    xsi_type: Optional[PolicyXsiType] = prop(
        None, repr=False, key=Jsonable.Prop.XSI_TYPE.value
    )
    name: Optional[str] = prop(None, repr=False)
    itg_id: Optional[int] = prop(None, converter=optional(int), repr=False)
    itg: Optional[str] = prop(None, repr=False)
    is_unique_active_in_itg: bool = prop(
        False, repr=False, key=Prop.UNIQUE_ACTIVE_IN_ITG.value
    )
    policy_targets: List[PolicyTarget] = prop(
        factory=list, repr=False, key=Prop.POLICY_TARGETS.value
    )
    installable_targets: List[InstallableTarget] = prop(
        factory=list, repr=False, flatify="installable_target"
    )


@propify
class BindingACL(Jsonable):
    class Prop(Enum):
        GLOBAL = "global"

    name: Optional[str] = prop(None, repr=False)
    is_global: Optional[bool] = prop(None, repr=False, key=Prop.GLOBAL.value)
    interfaces: List[Interface] = prop(factory=list, repr=False)
    ipv6_only: Optional[bool] = prop(None, repr=False)


@propify
class RuleBinding(Jsonable):
    class Prop(Enum):
        FROM_ZONE = "from_zone"
        TO_ZONE = "to_zone"

    class Direction(Enum):
        INBOUND = "inbound"
        OUTBOUND = "outbound"

    policy: Optional[BindingPolicy] = prop(None, repr=False)
    acl: Optional[BindingACL] = prop(None, repr=False)
    rule_count: int = prop(0)
    security_rule_count: int = prop(0)
    default: Optional[bool] = prop(None, repr=False)
    uid: Optional[str] = prop(None)
    direction: Optional[Direction] = prop(None)
    display_name: Optional[str] = prop(None)
    sub_policy_name: Optional[str] = prop(None, repr=False)
    is_shared: Optional[bool] = prop(None, repr=False)
    is_global: Optional[bool] = prop(None, repr=False)
    from_zones: List[Zone] = prop(factory=list, repr=False, key=Prop.FROM_ZONE.value)
    to_zones: List[Zone] = prop(factory=list, repr=False, key=Prop.TO_ZONE.value)
    ip_type: Optional[IPType] = prop(None, repr=False)


@propify
class SecurityRuleBase(Jsonable):
    class Prop(Enum):
        DISABLED = "disabled"
        BINDING = "binding"
        INSTALL = "install"

    class Type(Enum):
        RULE = "rule"

    id: int = prop(0, converter=int)
    device: Optional[Device] = prop(None)
    uid: Optional[str] = prop(None)
    order: Optional[int] = prop(None, converter=optional(int), repr=False)
    is_disabled: Optional[bool] = prop(None, repr=False, key=Prop.DISABLED.value)
    bindings: List[RuleBinding] = prop(factory=list, repr=False, key=Prop.BINDING.value)

    installs: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.INSTALL.value
    )


@propify
class NatRule(SecurityRuleBase):
    class Prop(Enum):
        AUTO_NAT = "autoNat"
        ORIG_SRC_NETWORK = "orig_src_network"
        ORIG_DST_NETWORK = "orig_dst_network"
        TRANSLATED_DST_NETWORK = "translated_dst_network"
        TRANSLATED_SRC_NETWORK = "translated_src_network"
        TRANSLATED_SERVICE = "translated_service"

    xsi_type: Optional[SecurityRuleType] = prop(
        None, repr=False, key=Jsonable.Prop.XSI_TYPE.value
    )
    is_auto_nat: bool = prop(None, repr=False, key=Prop.AUTO_NAT.value)
    egress_interface: List[Interface] = prop(factory=list, repr=False)
    orig_service: List[ObjectReference] = prop(factory=list, repr=False)

    install: List[str] = prop(factory=list, repr=False)
    enable_net4tonet6: Optional[bool] = prop(None, repr=False)
    enable_route_lookup: Optional[bool] = prop(None, repr=False)
    orig_src_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.ORIG_SRC_NETWORK.value
    )
    orig_dest_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.ORIG_DST_NETWORK.value
    )
    translated_dest_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.TRANSLATED_DST_NETWORK.value
    )
    translated_src_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.TRANSLATED_SRC_NETWORK.value
    )
    translated_services: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.TRANSLATED_SERVICE.value
    )
    type: Optional[SecurityRuleBase.Type] = prop(None, repr=False)


@propify
class RecordSet(Jsonable):
    class Prop(Enum):
        BUSINESS_OWNER_EMAIL = "businessOwnerEmail"
        BUSINESS_OWNER_NAME = "businessOwnerName"
        EXPIRATION_DATE = "expireDate"
        TICKET_CR = "ticketCr"
        TICKET_ORIGIN = "ticketOrigin"
        TICKET_STATUS = "ticketStatus"

    class TicketOrigin(Enum):
        SCW = "scw"

    business_owner_email: Optional[str] = prop(
        None, key=Prop.BUSINESS_OWNER_EMAIL.value
    )
    business_owner_name: Optional[str] = prop(None, key=Prop.BUSINESS_OWNER_NAME.value)
    expiration_date: Optional[datetime] = prop(
        None,
        key=Prop.EXPIRATION_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat() if d else None,
    )

    ticket_cr: str = prop("", key=Prop.TICKET_CR.value)
    ticket_origin: Optional[TicketOrigin] = prop(None, key=Prop.TICKET_ORIGIN.value)
    ticket_status: str = prop("", key=Prop.TICKET_STATUS.value)

    _exists_on_server: bool = prop(False, jsonify=False)


@propify
class Violation(Jsonable):
    name: str = prop("")
    type: str = prop("")


@propify
class SecurityPolicyViolation(Jsonable):
    class Severity(Enum):
        CRITICAL = "CRITICAL"
        HIGH = "HIGH"
        MEDIUM = "MEDIUM"
        LOW = "LOW"

    severity: Severity = prop()


class SecurityPolicyViolationType(Enum):
    SECURITY_POLICY = "SECURITY_POLICY"
    PCI = "PCI"
    SOX = "SOX"


@propify
class SecureAppApplication(Jsonable):
    app_name: str = prop("")
    app_owner: str = prop("")


def record_set_from_server(obj):
    record_set = RecordSet.kwargify(obj)
    record_set._exists_on_server = True

    return record_set


@propify
class Documentation(Jsonable):
    class Prop(Enum):
        RECORD_SETS = "record_set"
        SECURE_APP_APPLICATIONS = "secure_app_application"

    class Shadowed(Enum):
        Fully = "Fully shadowed"
        Not = "Not shadowed"
        NotApplicable = "N/A"
        Empty = ""

    class CertificationStatus(Enum):
        CERTIFIED = "CERTIFIED"
        DECERTIFIED = "DECERTIFIED"

    comment: Optional[str] = prop(None)
    legacy_rule: Optional[bool] = prop(None)
    stealth_rule: Optional[bool] = prop(None)
    shadowed: Optional[Shadowed] = prop(None)
    record_sets: List[RecordSet] = prop(
        factory=list, key=Prop.RECORD_SETS.value, kwargify=record_set_from_server
    )
    last_modified: Optional[str] = prop(None)
    last_hit: Optional[str] = prop(None)
    permissiveness_level: Optional[str] = prop(None)

    violations: List[Violation] = prop(factory=list, flatify="violation")
    certification_status: Optional[CertificationStatus] = prop(None)
    certification_date: Optional[datetime] = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )

    certification_expiration_date: Optional[datetime] = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )

    secure_app_applications: List[SecureAppApplication] = prop(
        factory=list, key=Prop.SECURE_APP_APPLICATIONS.value
    )

    tech_owner: Optional[str] = prop(None)

    @property
    def _json(self) -> dict:
        if self._json_override is not None:
            return self._json_override
        obj = deepcopy(self)
        obj.record_sets = [rs for rs in obj.record_sets if not rs._exists_on_server]

        return jsonify(obj)


@propify
class SecurityRule(SecurityRuleBase):
    class Meta(Enum):
        ROOT = "rule"

    class Prop(Enum):
        IMPLICIT = "implicit"
        EXTERNAL = "external"
        TRACK = "track"
        AUTHENTICATION_STATUS = "authentication_status"
        SRC_NETWORK = "src_network"
        DST_NETWORK = "dst_network"
        SRC_SERVICE = "src_service"
        DST_SERVICE = "dst_service"
        APPLICATION = "application"
        TIME = "time"
        VPN = "vpn"
        OPTION = "option"
        URL_CATEGORY = "url_category"
        SRC_ZONE = "src_zone"
        DST_ZONE = "dst_zone"
        SRC_ZONE_ANY = "src_zone_any"
        DST_ZONE_ANY = "dst_zone_any"
        ADDITIONAL_PARAMETER = "additional_parameter"
        AUTHENTICATION_RULE = "authentication_rule"
        USERS = "users"
        ASSOCIATED_NAT_RULE = "associatedNatRule"
        SUB_POLICY_GLOBAL = "sub_policy_global"
        SUB_POLICY_SHARED = "sub_policy_shared"

    class GlobalLocation(Enum):
        AFTER = "AFTER"
        BEFORE = "BEFORE"
        MIDDLE = "MIDDLE"

    class Action(Enum):
        ACCEPT = "Accept"
        CONTINUE = "Continue"
        DENY = "Deny"
        DROP = "Drop"
        INLINE_LAYER = "Inline Layer"
        REJECT = "Reject"

    class Option(Enum):
        LOG = "LOG"
        LOG_FORWARDING = "LOG_FORWARDING"
        LOG_SESSION = "LOG_SESSION"
        RETURNING_TRAFFIC = "RETURNING_TRAFFIC"

    class RuleType(Enum):
        INTERZONE = "interzone"
        INTRAZONE = "intrazone"
        UNIVERSAL = "universal"

    xsi_type: Optional[SecurityRuleType] = prop(
        None, repr=False, key=Jsonable.Prop.XSI_TYPE.value
    )
    cp_uid: Optional[str] = prop(None, repr=False)
    name: Optional[str] = prop(None, repr=False)
    comment: Optional[str] = prop(None, repr=False)
    action: Optional[Action] = prop(None)
    is_implicit: Optional[bool] = prop(None, repr=False, key=Prop.IMPLICIT.value)
    is_external: Optional[bool] = prop(None, repr=False, key=Prop.EXTERNAL.value)
    global_location: Optional[GlobalLocation] = prop(False, repr=False)
    acceleration_breaker: Optional[bool] = prop(None, repr=False)
    is_authentication_rule: bool = prop(
        False, repr=False, key=Prop.AUTHENTICATION_RULE.value
    )
    rule_number: Optional[int] = prop(None, converter=optional(int), repr=False)
    track: Optional[Track] = prop(None, repr=False)
    options: List[Option] = prop(factory=list, repr=False, key=Prop.OPTION.value)
    src_networks_negated: Optional[bool] = prop(None, repr=False)
    dest_networks_negated: Optional[bool] = prop(None, repr=False)
    src_services_negated: Optional[bool] = prop(None, repr=False)
    dest_services_negated: Optional[bool] = prop(None, repr=False)
    src_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.SRC_NETWORK.value
    )
    dest_networks: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.DST_NETWORK.value
    )
    src_services: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.SRC_SERVICE.value
    )
    dest_services: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.DST_SERVICE.value
    )
    applications: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.APPLICATION.value
    )
    times: List[ObjectReference] = prop(factory=list, repr=False, key=Prop.TIME.value)

    vpns: List[ObjectReference] = prop(factory=list, repr=False, key=Prop.VPN.value)

    users: List[ObjectReference] = prop(factory=list, repr=False)
    user_access: List[ObjectReference] = prop(factory=list, repr=False)
    url_categories: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.URL_CATEGORY.value
    )
    rule_type: Optional[RuleType] = prop(None, repr=False)
    src_zones: List[str] = prop(factory=list, repr=False, key=Prop.SRC_ZONE.value)
    dest_zones: List[str] = prop(factory=list, repr=False, key=Prop.DST_ZONE.value)
    is_src_zone_any: Optional[bool] = prop(
        None, repr=False, key=Prop.SRC_ZONE_ANY.value
    )
    is_dest_zone_any: Optional[bool] = prop(
        None, repr=False, key=Prop.DST_ZONE_ANY.value
    )
    rule_location: Optional[str] = prop(None, repr=False)
    rule_location_display: Optional[str] = prop(None, repr=False)
    additional_parameters: List[ObjectReference] = prop(
        factory=list, repr=False, key=Prop.ADDITIONAL_PARAMETER.value
    )
    type: Optional[SecurityRuleBase.Type] = prop(None, repr=False)

    ip_type: Optional[IPType] = prop(None, repr=False)
    fmg_from_zone: Optional[str] = prop(None, repr=False)
    fmg_to_zone: Optional[str] = prop(None, repr=False)
    associated_nat_rule: Optional[NatRule] = prop(
        None, repr=False, key=Prop.ASSOCIATED_NAT_RULE.value
    )
    textual_rep: Optional[str] = prop(None, repr=False)

    priority: Optional[str] = prop(None, repr=False)
    sub_policy: Optional[str] = prop(None, repr=False)
    sub_policy_uid: Optional[str] = prop(None, repr=False)
    is_sub_policy_global: bool = prop(
        False, repr=False, key=Prop.SUB_POLICY_GLOBAL.value
    )
    is_sub_policy_shared: bool = prop(
        False, repr=False, key=Prop.SUB_POLICY_SHARED.value
    )
    documentation: Optional[Documentation] = prop(None)


@propify
class RuleLastUsage(Jsonable):
    rule_uid: str = prop("", key="rule_UID")
    users: dict = prop(None)
    applications: dict = prop(None)
    rule_last_hit: Optional[datetime] = prop(
        None,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat() if d else None,
        repr=lambda v: v.isoformat() if v else "None",
    )



import pytest
import json
import responses

from pytos2.securetrack.generic_interface_customer import GenericInterfaceCustomerTag
from pytos2.utils import get_api_node


class TestGenericInterfaceCustomer:
    device = json.load(
        open("tests/securetrack/json/generic_interface_customers/device-5.json")
    )
    interfaces = device["InterfaceCustomerTags"]

    interface_customer = json.load(
        open("tests/securetrack/json/generic_interface_customers/int-cust-74.json")
    )
    interface = interface_customer["InterfaceCustomerTag"]

    @responses.activate
    def test_add_generic_interface_customers_tag_200(self, st):
        """Add one or multiple interface customers"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=201,
        )

        """Add single generic interface customer tag"""
        newRoute = st.add_generic_interface_customer_tag(
            generic=self.interface["generic"],
            device=self.interface["deviceId"],
            interface_name=self.interface["interfaceName"],
            customer_id=self.interface["customerId"],
        )
        assert newRoute is None

    @responses.activate
    def test_add_generic_interface_customer_tags_400(self, st):
        """Add one or multiple interface customer tags"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=400,
        )

        """Add single interface customer tag"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customer_tag(
                generic=self.interface["generic"],
                device=self.interface["deviceId"],
                interface_name=self.interface["interfaceName"],
                customer_id=self.interface["customerId"],
            )
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_interface_customer_tags_404(self, st):
        """Add one or multiple interface customers"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=404,
        )

        """Add single interface customer"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_interface_customer_tag(
                generic=self.interface["generic"],
                device=self.interface["deviceId"],
                interface_name=self.interface["interfaceName"],
                customer_id=self.interface["customerId"],
            )
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interface_customer_tag(
        self, st, generic_interface_customer_tag_mock
    ):
        """Get interface by interface id"""
        mockRes = get_api_node(self.interface_customer, "InterfaceCustomerTag")
        interface = GenericInterfaceCustomerTag.kwargify(mockRes)
        interface_customerByInt = st.get_generic_interface_customer_tag(74)
        assert interface_customerByInt == interface

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customer_tag(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_interface_customer_tags(
        self, st, generic_interface_customer_tag_mock
    ):
        """Get interfaces by device id"""
        interfaces = [
            GenericInterfaceCustomerTag.kwargify(d)
            for d in get_api_node(self.device, "InterfaceCustomerTags", listify=True)
        ]
        interface_customersByInt = st.get_generic_interface_customer_tags(
            5, generic=False
        )
        assert interface_customersByInt == interfaces

        with pytest.raises(ValueError) as exception:
            st.get_generic_interface_customer_tags(404, generic=True)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_interface_customers_200(self, st):
        """Update one or multiple interface customers"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=200,
        )

        """Update single interface customer"""
        interface_to_update = GenericInterfaceCustomerTag.kwargify(self.interface)
        newRoute = st.update_generic_interface_customer_tag(interface_to_update)
        assert newRoute is None

    @responses.activate
    def test_update_generic_interface_customers_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            interface_to_update = GenericInterfaceCustomerTag.kwargify(self.interface)
            st.update_generic_interface_customer_tag(interface_to_update)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_interface_customers_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/interfacecustomer",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            interface_to_update = GenericInterfaceCustomerTag.kwargify(self.interface)
            st.update_generic_interface_customer_tag(interface_to_update)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_interface_customer_tag(
        self, st, generic_interface_customer_tag_mock
    ):
        """Delete interface by interface id"""
        interface_customerByInt = st.delete_generic_interface_customer_tag(74)
        assert interface_customerByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customer_tag(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_interface_customer_tags(
        self, st, generic_interface_customer_tag_mock
    ):
        """Delete interfaces by device id"""
        interface_customersByInt = st.delete_generic_interface_customer_tags(5)
        assert interface_customersByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_interface_customer_tags(404)
        assert "Not Found" in str(exception.value)



import pytest
import json
import responses

from pytos2.securechange.workflow import (
    BasicWorkflowInfo,
    WorkflowType,
    FullWorkflow,
    FullARWorkflow,
)


class TestWorkflows:
    @responses.activate
    def test_get_workflows(self, workflows_mock, scw):
        workflows = scw.get_workflows()
        assert workflows[0].id == 588
        assert workflows[0].name == "Contoso"
        assert workflows[0].description == ""
        assert workflows[0].type == WorkflowType.ACCESS_REQUEST
        unique_types = sorted(set([w.type.value for w in workflows]))
        assert unique_types == [
            WorkflowType.ACCESS_REQUEST.value,
            WorkflowType.GENERIC.value,
        ]

        workflows = scw.get_workflows(workflow_type=WorkflowType.ACCESS_REQUEST)
        assert workflows[0].id == 588
        assert workflows[0].name == "Contoso"
        assert workflows[0].description == ""
        assert workflows[0].type == WorkflowType.ACCESS_REQUEST
        unique_types = sorted(set([w.type.value for w in workflows]))
        assert unique_types == [WorkflowType.ACCESS_REQUEST.value]

    @responses.activate
    def test_get_workflow(self, workflows_mock, scw):
        workflow = scw.get_workflow(586)
        assert isinstance(workflow, FullWorkflow)
        assert isinstance(workflow, FullARWorkflow)
        assert workflow.id == 586
        assert workflow.name == "Fully automated Firewall change request"

        workflow = scw.get_workflow(
            workflow_name="Fully automated Firewall change request"
        )
        assert isinstance(workflow, FullWorkflow)
        assert workflow.id == 586
        assert workflow.name == "Fully automated Firewall change request"

        workflow = scw.get_workflow(600)
        assert isinstance(workflow, FullWorkflow)
        assert type(workflow) is not FullARWorkflow
        assert workflow.id == 600
        assert workflow.name == "Generic Workflow"

        workflow = scw.get_workflow(workflow_name="Generic Workflow")
        assert isinstance(workflow, FullWorkflow)
        assert type(workflow) is not FullARWorkflow
        assert workflow.id == 600
        assert workflow.name == "Generic Workflow"



from typing import Optional, List, Union
from enum import Enum

from pytos2.utils import propify, prop
from pytos2.models import Jsonable

# Prevents breaking changes
from .network_object import (
    classify_object_type,
    NetworkObject,
    HostObject,
    HostNetworkObject,
    SubnetObject,
    RangeObject,
    GroupObject,
)


@propify
class ServiceObject(Jsonable):
    class Prop(Enum):
        XSI_TYPE = "@xsi.type"
        UID = "uid"
        GLOBAL = "global"

    class ServiceObjectXsiType(Enum):
        TRANSPORT_SERVICE_TYPE = "transport_service"
        IP_SERVICE_TYPE = "ip_service"
        ICMP_SERVICE_TYPE = "icmp_service"
        SERVICE_GROUP_TYPE = "service_group"
        RPC_SERVICE_TYPE = "rpc_service"
        ANY_APPLICATION_TYPE = "any_application"
        ANY_SERVICE_TYPE = "any_service"

    class Referenced(Enum):
        FALSE = "FALSE"
        SELF_TABLE_ONLY = "SELF_TABLE_ONLY"
        UNKNOWN = "UNKNOWN"
        FROM_RULE = "FROM_RULE"
        FROM_NON_RULE_OBJECTS = "FROM_NON_RULE_OBJECTS"
        TRUE = "TRUE"

    xsi_type: ServiceObjectXsiType = prop(ServiceObjectXsiType.TRANSPORT_SERVICE_TYPE)

    min_value_source: Optional[int] = prop(None)
    max_value_source: Optional[int] = prop(None)

    uid: Optional[Union[str, int]] = prop(
        None, converter=str, repr=False, key=Prop.UID.value
    )
    name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)
    origin: Optional[str] = prop(None, repr=False)
    is_global: Optional[bool] = prop(None, key=Prop.GLOBAL.value, cmp=False, repr=False)
    implicit: Optional[bool] = prop(None, repr=False)
    shared: Optional[bool] = prop(None, repr=False)
    timeout: Optional[int] = prop(None, repr=False)

    comment: Optional[str] = prop(None, repr=False)
    version_id: Optional[int] = prop(None, repr=False)
    referenced: Optional[Referenced] = prop(None, repr=False)
    type_on_device: Optional[str] = prop(None, repr=False)

    negate: Optional[bool] = prop(None, repr=False)
    match_for_any: Optional[bool] = prop(None, repr=False)

    @classmethod
    def from_securetrack(cls, obj):
        return _from_securetrack(obj)


def _from_securetrack(obj):
    from pytos2.securetrack.service_object import (
        TCPServiceObject,
        UDPServiceObject,
        PortServiceObject,
        ICMPServiceObject,
        ICMPV6ServiceObject,
        ICMPV6IPServiceObject,
        ServiceGroup,
        DCERPCService,
        RPCServiceObject,
        OtherIPServiceObject,
        AnyIPServiceObject,
    )

    def _mv_key(j, src, dest):
        if src not in j:
            return

        j[dest] = j[src]
        del j[src]

    j = dict(obj._json)

    if isinstance(obj, (TCPServiceObject, UDPServiceObject, PortServiceObject)):
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.TRANSPORT_SERVICE_TYPE.value
        _mv_key(j, "min", "min_port")
        _mv_key(j, "max", "max_port")

        return TransportService.kwargify(j)
    elif isinstance(
        obj, (ICMPServiceObject, ICMPV6ServiceObject, ICMPV6IPServiceObject)
    ):
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.ICMP_SERVICE_TYPE.value
        _mv_key(j, "min", "min_icmp_type")
        _mv_key(j, "max", "max_icmp_type")
        return ICMPService.kwargify(j)
    elif isinstance(obj, ServiceGroup):
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.SERVICE_GROUP_TYPE.value
        return ServiceGroup.kwargify(j)
    elif isinstance(obj, (DCERPCService, RPCServiceObject)):
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.RPC_SERVICE_TYPE.value
        return ServiceObject.kwargify(j)
    elif isinstance(obj, (OtherIPServiceObject, AnyIPServiceObject)):
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.IP_SERVICE_TYPE.value
        return ServiceObject.kwargify(j)
    else:
        j["@xsi.type"] = ServiceObject.ServiceObjectXsiType.TRANSPORT_SERVICE_TYPE.value
        return TransportService.kwargify(j)


@propify
class TransportService(ServiceObject):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.TRANSPORT_SERVICE_TYPE
    )
    protocol: Optional[int] = prop(None)
    min_port: Optional[int] = prop(None)
    max_port: Optional[int] = prop(None)


@propify
class IPService(ServiceObject):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.IP_SERVICE_TYPE
    )
    min_protocol: Optional[int] = prop(None)
    max_protocol: Optional[int] = prop(None)


@propify
class AnyService(IPService):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.ANY_SERVICE_TYPE
    )


@propify
class ICMPService(ServiceObject):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.ICMP_SERVICE_TYPE
    )
    min_icmp_type: Optional[int] = prop(None)
    max_icmp_type: Optional[int] = prop(None)
    icmp_code: Optional[int] = prop(None)

    management_domain: Optional[str] = prop(None)


@propify
class AnyApplicationService(ServiceObject):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.ANY_APPLICATION_TYPE
    )
    protocol: Optional[int] = prop(None)
    min_port: Optional[int] = prop(None)
    max_port: Optional[int] = prop(None)


@propify
class RPCService(ServiceObject):
    xsi_type: ServiceObject.ServiceObjectXsiType = prop(
        ServiceObject.ServiceObjectXsiType.RPC_SERVICE_TYPE
    )


"""
@propify
class NetworkObject(Jsonable):
    class Prop(Enum):
        XSI_TYPE = "@xsi.type"
        OBJECT_UID = "object_UID"
        GLOBAL = "global"


    uid: Optional[str] = prop(None, converter=str, repr=False)
    name: Optional[str] = prop(None, repr=False)
    display_name: Optional[str] = prop(None)
    origin: Optional[str] = prop(None, repr=False)
    is_global: Optional[bool] = prop(None, key=Prop.GLOBAL.value, cmp=False, repr=False)

    implicit: Optional[bool] = prop(None, repr=False)
    shared: Optional[bool] = prop(None, repr=False)
    ip_type: Optional[str] = prop(None)
    installable_target: Optional[bool] = prop(None, repr=False)
    group_ids: Optional[str] = prop(None, repr=False)  # belone to group
"""


def classify_service_type(service_object):
    if (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.TRANSPORT_SERVICE_TYPE.value
    ):
        return TransportService.kwargify(service_object)
    elif (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.IP_SERVICE_TYPE.value
    ):
        return IPService.kwargify(service_object)
    elif (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.ICMP_SERVICE_TYPE.value
    ):
        return ICMPService.kwargify(service_object)
    elif (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.SERVICE_GROUP_TYPE.value
    ):
        return ServiceGroup.kwargify(service_object)
    elif (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.ANY_APPLICATION_TYPE.value
    ):
        return AnyApplicationService.kwargify(service_object)
    elif (
        service_object["@xsi.type"]
        == ServiceObject.ServiceObjectXsiType.ANY_SERVICE_TYPE.value
    ):
        return AnyService.kwargify(service_object)
    else:
        return ServiceObject.kwargify(service_object)


@propify
class ServiceGroup(ServiceObject):
    class Prop(Enum):
        MEMBER = "member"

    members: List[Union[TransportService, IPService, ICMPService]] = prop(
        factory=list, repr=False, key=Prop.MEMBER.value, kwargify=classify_service_type
    )



from enum import Enum
from datetime import date, datetime
from typing import Optional, List

from pytos2.models import Jsonable, ObjectReference, FQDNIp
from pytos2.utils import propify, prop, safe_iso8601_date


def classify_application(obj):
    return get_application_class(obj).kwargify(obj)


def get_application_class(obj):
    if "@xsi.type" in obj:
        mapping = {
            "singleApplicationDTO": SingleApplication,
            "applicationGroupDTO": ApplicationGroup,
        }
        return mapping.get(obj["@xsi.type"], Application)
    else:
        return Application


@propify
class Link(Jsonable):
    href: str = prop("", key="@href")


@propify
class Application(Jsonable):
    name: str = prop("")
    display_name: str = prop("")
    class_name: str = prop("")
    type: str = prop("")
    comment: str = prop("")
    uid: str = prop("")
    overrides: bool = prop(None, repr=False)
    management_domain: str = prop("")
    global_device: bool = prop(None, repr=False)
    device_id: int = prop(None)
    application_id: int = prop("")
    device_name: str = prop("", key="deviceName")


@propify
class SingleApplication(Application):
    services: List[ObjectReference] = prop(factory=list, key="service")


@propify
class ApplicationGroup(Application):
    applications: List[ObjectReference] = prop(factory=list, key="application")



import json

import pytest
import responses
from netaddr import IPAddress, IPNetwork

from requests.exceptions import HTTPError

# from result import Result
from typing import Iterator

# from pytos2.securetrack.entrypoint import St
from pytos2.api import Pager
from pytos2.securetrack.rule import BindingPolicy, SecurityRule
from pytos2.securetrack.generic_device import GenericDevice
from pytos2.securetrack.topology import TopologySyncStatus
from pytos2.securetrack.zone import ZoneReference, ZoneEntry
from pytos2.securetrack.revision import Revision
from pytos2.securetrack import St
from pytos2.securetrack.device import Device
from pytos2.securetrack.network_object import HostNetworkObject, BasicNetworkObject
from pytos2.securetrack.service_object import (
    Service,
    OtherIPServiceObject,
    UDPServiceObject,
    AnyIPServiceObject,
)


class TestClient(object):
    def test_with_args(self, st_api):
        assert st_api.hostname == "198.18.0.1"
        assert st_api.username == "username"
        assert st_api.password == "password"

    def test_with_missing_args(self):
        with pytest.raises(ValueError) as e:
            St()
        assert "hostname argument must be provided" in str(e.value)


class TestClientMethods:
    @responses.activate
    def test_get_device(self, devices_mock, st):
        device = st.get_device(8)
        assert device.id == 8
        device = st.get_device("ASAv")
        assert device.name == "ASAv"

    @responses.activate
    def test_get_devices(self, devices_mock, st):
        filter_dict = {"vendor": "Checkpoint", "parent_id": 357}

        devices = st.get_devices(cache=False, filter=filter_dict)
        assert isinstance(st.get_devices(cache=False), list)
        assert st._devices_cache.is_empty()
        assert isinstance(st.get_devices(cache=False, filter=filter_dict), list)
        assert st._devices_cache.is_empty()
        assert isinstance(st.get_devices(), list)
        assert len(devices) == 2
        assert isinstance(st.get_devices(cache=True, filter=filter_dict), list)
        assert st._devices_cache.is_empty() is False
        assert isinstance(st.get_devices(), list)
        assert len(devices) == 2

    @responses.activate
    def test_get_network_objects(self, network_objects_mock, st):
        assert isinstance(st.get_network_objects(device=20), list)
        with pytest.raises(ValueError):
            st.get_network_objects(device=10000)

        device = st.get_device(20)
        network_objs = device.get_network_objects()
        assert isinstance(network_objs, list)
        assert isinstance(network_objs[0], BasicNetworkObject)

        assert isinstance(st.get_network_objects(device=20, cache=False), Pager)

        assert isinstance(
            st.get_network_objects(device=731, add_parent_objects=False), list
        )

        assert isinstance(
            st.get_network_objects(device=731, add_parent_objects=False, cache=False),
            Pager,
        )

    @responses.activate
    def test_get_device_no_cache(self, st_no_cache):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/devices/8\?show_license=false\&show_os_version=false",
            json=json.load(open("tests/securetrack/json/devices/8.json")),
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/devices/8",
            json=json.load(open("tests/securetrack/json/devices/8.json")),
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/devices?name=ASAv",
            json=json.load(open("tests/securetrack/json/devices/ASAv.json")),
        )
        device = st_no_cache.get_device(8)
        assert device.id == 8
        device = st_no_cache.get_device("ASAv")
        assert device.name == "ASAv"

    @responses.activate
    def test_get_rules_for_device(self, devices_for_rule_test_mock, rules_mock, st):
        with pytest.raises(ValueError):
            st.get_rules(device=1, revision=1)

        rules = st.get_rules()

        for rule in rules:
            assert isinstance(rule, SecurityRule)

        st._device_rules_dict = {}
        rules = st.get_rules(rule_uid="{79985494-C73E-11E4-BD2B-7F000001F4F4}")

        for rule in rules:
            assert isinstance(rule, SecurityRule)

        # Test cache if-branch
        rules = st.get_rules(device=20)
        for rule in rules:
            assert isinstance(rule, SecurityRule)

        rules = st.get_rules(
            device=20, rule_uid="{79985494-C73E-11E4-BD2B-7F000001F4F4}"
        )

        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        assert rule.name == "Access to exchange server"

        # Test non-cached if-branch for rule uid
        st._device_rules_dict = {}
        rules = st.get_rules(
            device=20, rule_uid="{79985494-C73E-11E4-BD2B-7F000001F4F4}"
        )

        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        assert rule.name == "Access to exchange server"

        rules = st.get_rules(device=20, uid="{3A1BA062-6B19-4C97-8F18-79CBA9EF0AA6}")
        rule = rules[0]

        assert isinstance(rule, SecurityRule)

        with pytest.raises(ValueError):
            st.get_rules(device=400000)

    @responses.activate
    def test_get_rules_for_revision(self, revisions_mock, rules_mock, st):
        rules = st.get_rules(revision=2285)
        assert isinstance(rules, list)

        rules = st.get_rules(
            revision=2285, rule_uid="{ddf6dfbd-2080-45c9-8476-89a87d0ae5a8}"
        )

        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        assert rule.id == 1617

        rules = st.get_rules(revision=2285, uid="Datacenter_access_in_@_8")
        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        with pytest.raises(ValueError):
            st.get_rules(revision=400000)

        with pytest.raises(ValueError):
            st.get_rules(revision=2226)

        with pytest.raises(ValueError):
            st.get_rules(revision=2226, cache=False)

    @responses.activate
    def test_get_rules_from_device_no_cache(
        self, st_no_cache, devices_mock, rules_mock
    ):
        with pytest.raises(ValueError):
            st_no_cache.get_rules()
            # given conftest.py, will get 500 error on device 10

        rules = st_no_cache.get_rules(device=20)
        assert isinstance(rules[0], SecurityRule)

        rules = st_no_cache.get_rules(
            device=20, rule_uid="{79985494-C73E-11E4-BD2B-7F000001F4F4}"
        )

        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        assert rule.id == 4444

        rules = st_no_cache.get_rules(
            device=20, uid="{3A1BA062-6B19-4C97-8F18-79CBA9EF0AA6}"
        )
        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        with pytest.raises(HTTPError):
            st_no_cache.get_rules(device=400000)

    @responses.activate
    def test_get_rules_from_revision_no_cache(
        self, st_no_cache, revisions_mock, rules_mock
    ):
        rules = st_no_cache.get_rules(revision=2285)
        assert isinstance(rules[0], SecurityRule)

        rules = st_no_cache.get_rules(
            revision=2285, rule_uid="{ddf6dfbd-2080-45c9-8476-89a87d0ae5a8}"
        )
        rule = rules[0]
        assert isinstance(rule, SecurityRule)
        assert rule.id == 1617

        rules = st_no_cache.get_rules(revision=2285, uid="Datacenter_access_in_@_8")
        rule = rules[0]
        assert isinstance(rule, SecurityRule)

        with pytest.raises(ValueError):
            st_no_cache.get_rules(revision=400000)
        with pytest.raises(ValueError):
            st_no_cache.get_rules(revision=2226)

    @responses.activate
    def test_search_rules(self, devices_mock, search_rules_mock, st):
        rules = st.search_rules(devices=105)
        # rules = [r for r in rules]

        assert len(rules) == 3230

    @responses.activate
    def test_get_rules_on_open_tickets(self, search_rules_on_open_tickets_mock, st):
        rules = st.get_rules_on_open_tickets()

        assert len(rules) == 20

    @responses.activate
    def test_get_revision(self, devices_mock, revisions_mock, st):
        revision = st.get_revision(revision=2285)
        assert isinstance(revision, Revision)
        assert revision.revision_id == 135

        # Second time should run through cache
        revision = st.get_revision(revision=2285)
        assert isinstance(revision, Revision)

        with pytest.raises(ValueError):
            st.get_revision(revision=400000)

    @responses.activate
    def test_get_revisions(self, devices_mock, revisions_mock, st):
        revisions = st.get_revisions(device=8)
        assert isinstance(revisions, list)

        assert isinstance(revisions[0], Revision)
        assert revisions[0].revision_id == 135

        revisions = st.get_revisions(device="ASAv")
        assert isinstance(revisions, list)
        with pytest.raises(ValueError):
            st.get_revisions(device="RTR33333")

        device = st.get_device(identifier=8)
        assert isinstance(device.get_revisions(), list)

    @responses.activate
    def test_get_latest_revision(self, devices_mock, revisions_mock, st):
        revision = st.get_latest_revision(device=8)
        assert revision.revision_id == 135
        with pytest.raises(HTTPError):
            st.get_latest_revision(device=100000)

    @responses.activate
    def test_get_revisions_no_cache(self, devices_mock, revisions_mock, st_no_cache):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/devices/8/revisions",
            json=json.load(open("tests/securetrack/json/revisions/device-8.json")),
        )

        revisions = st_no_cache.get_revisions(device=8, cache=False)
        assert isinstance(revisions, list)

        with pytest.raises(HTTPError):
            st_no_cache.get_revisions(device=100000, cache=False)

    @responses.activate
    def test_get_revision_no_cache(self, devices_mock, revisions_mock, st_no_cache):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/revisions/2285",
            json=json.load(open("tests/securetrack/json/revisions/revision-2285.json")),
        )

        revision = st_no_cache.get_revision(2285, cache=False)
        assert isinstance(revision, Revision)

        with pytest.raises(ValueError):
            st_no_cache.get_revision(400000, cache=False)

    @responses.activate
    def test_get_network_object_by_name(self, network_objects_mock, st):
        network_object = st.get_network_object("Subnet_10.3.3.0", 157)
        assert network_object["name"] == "Subnet_10.3.3.0"

    @responses.activate
    def test_get_network_object_by_uid(self, network_objects_mock, st):
        network_object = st.get_network_object(
            uid="{b63f31eb-d8ab-410c-b526-20a48ff9dbd2}"
        )
        assert network_object["name"] == "Subnet_10.2.2.0"

    @responses.activate
    def test_get_network_object_by_uid_cached(self, network_objects_mock, st):
        st.get_network_object(device=181, name="1.1.1.1")
        st.get_network_object(device=174, name="1.1.1.1")
        network_object = st.get_network_object(
            uid="{b17b4120-07d4-0875-d4b2-06227767ff57}", device=174
        )
        assert network_object["name"] == "1.1.1.1"

    @responses.activate
    def test_get_multiple_network_object_by_uid_error(self, network_objects_mock, st):
        with pytest.raises(AssertionError):
            st.get_network_object(uid="{b17b4120-07d4-0875-d4b2-06227767ff57}")

    @responses.activate
    def test_get_multiple_network_object_by_uid(self, network_objects_mock, st):
        network_object = st.get_network_object(
            uid="b17b4120-07d4-0875-d4b2-06227767ff57", device=181
        )
        assert network_object["name"] == network_object.name == "1.1.1.1"

    @responses.activate
    def test_get_multiple_network_object_by_uid_with_device(
        self, network_objects_mock, st
    ):
        network_object = st.get_network_object(
            uid="b17b4120-07d4-0875-d4b2-06227767ff57", device=174
        )
        assert isinstance(network_object, HostNetworkObject)
        assert network_object["name"] == network_object.name == "1.1.1.1"

    @responses.activate
    def test_get_network_object_args(self, network_objects_mock, st):
        with pytest.raises(ValueError):
            st.get_network_object(device=174)

    @responses.activate
    def test_get_multiple_network_object_not_found(self, network_objects_mock, st):
        network_object = st.get_network_object(
            uid="b17b4120-07d4-0875-d4b2-06227767ff57", device=8
        )
        assert network_object is None

    @responses.activate
    def test_prime_services_cache(self, st, services_mock):
        assert len(st._services_by_device_id_by_name.get(1, {})) == 0

        st._prime_services_cache(device_id=1)

        assert len(st._services_by_device_id_by_name[1]) > 0

    @responses.activate
    def test_get_services(self, st, services_mock, devices_mock):
        services = st.get_services(device=1)
        assert isinstance(services, list)

        assert len(services) == 420

        assert services[0].id == 2878000
        assert services[0].name == "902 (tcp)"
        assert services[0].display_name == "902"
        assert services[0].class_name == Service.ClassName.TCP_SERVICE
        assert services[0].type == Service.Type.TCP_SERVICE
        assert not services[0].is_global
        assert services[0].comment == ""
        assert services[0].min_port == 902
        assert services[0].max_port == 902

        with pytest.raises(ValueError):
            st.get_services(device=5000000)

    # test cached service
    @responses.activate
    def test_get_service_preprimed(self, st, services_mock, devices_mock):
        # Prime the cache
        st.get_services(device=1)

        # UID case mismatch? This should still work.
        service = st.get_service(uid="{97aeb369-9aea-11d5-bd16-0090272ccb30}")
        assert isinstance(service, AnyIPServiceObject)
        assert service.id == 2877997
        assert service.name == "Any"
        assert not service.is_global
        assert service.class_name == Service.ClassName.ANY_OBJECT
        assert service.type == Service.Type.IP_SERVICE
        assert service.min_port == 0
        assert service.max_port == 255

    # test uncached service (from device 2 or 3)
    @responses.activate
    def test_get_service(self, st, services_mock, devices_mock):
        with pytest.raises(ValueError):
            service = st.get_service(name="aol")

        # No cache priming this time
        with pytest.raises(AssertionError):
            st.get_service(uid="{3fbd8116-3801-4bee-8593-3cbf999da671}")
        # Test once with caching of this particular UID
        service = st.get_service(uid="{3fbd8116-3801-4bee-8593-3cbf999da671}", device=1)
        assert isinstance(service, OtherIPServiceObject)

        assert service.id == 2878412
        assert service.name == "ah"
        assert service.comment == "Protocol: ah (51)"

        service = st.get_service(
            uid="{3fbd8116-3801-4bee-8593-3cbf999da671}", device=10
        )
        assert service is None

        st._services_by_device_id_by_name[1] = {}
        st._services_by_uid = {}

        # And once without.
        service = st.get_service(uid="{3fbd8116-3801-4bee-8593-3cbf999da671}", device=1)
        assert isinstance(service, OtherIPServiceObject)
        assert service.id == 2878412
        assert service.name == "ah"
        assert service.comment == "Protocol: ah (51)"

        service = st.get_service(device=1, name="aol (udp)")
        assert isinstance(service, UDPServiceObject)

        assert service.id == 2878007
        assert service.type == Service.Type.UDP_SERVICE
        assert service.comment == "America On-line"

        with pytest.raises(ValueError):
            st.get_service(device=50000000, name="aol (udp)")

    @responses.activate
    def test_get_network_object(self, st, network_objects_mock, devices_mock):
        with pytest.raises(ValueError):
            st.get_network_object(device=100000000, uid="nonexistent_uid")

    @responses.activate
    def test_get_device_policies(self, st, device_policies_mock, devices_mock):
        policies = st.get_device_policies(device=20)

        assert isinstance(policies, list)

        policy = policies[0]
        assert policy.name == "Toronto"
        assert isinstance(policy, BindingPolicy)

        policy = policies[1]

        assert policy.name == "Standard"
        assert isinstance(policy, BindingPolicy)

        with pytest.raises(IndexError):
            _ = policies[2]

        with pytest.raises(HTTPError):
            st.get_device_policies(device=400000)

        policies = st.get_device_policies(device="CP SMC")

        policy = policies[0]
        assert isinstance(policy, BindingPolicy)
        assert policy.name == "Toronto"

        with pytest.raises(ValueError):
            st.get_device_policies(device="NONEXISTENT_ABCDEFGH")

    @responses.activate
    def test_get_device_policy(self, st, device_policies_mock, devices_mock):
        policy = st.get_device_policy(device=20, policy="Standard")
        assert isinstance(policy, BindingPolicy)
        assert policy.name == "Standard"

        with pytest.raises(ValueError):
            st.get_device_policy(device=20, policy="Nonexistent")

        with pytest.raises(ValueError):
            st.get_device_policy(device="NONEXISTENT_ABCDEFGH", policy="Nonexistent")

    @responses.activate
    def test_get_generic_devices_error(self, st, generic_devices_getter_error):
        with pytest.raises(ValueError):
            st.get_generic_devices(cache=True)

        with pytest.raises(ValueError):
            st.get_generic_devices(cache=False)

    @responses.activate
    def test_get_generic_devices(self, st, generic_devices_mock):
        devices_list = st.get_generic_devices()
        for d in devices_list:
            assert isinstance(d, GenericDevice)

        assert devices_list[0].name == "Generic02"

        devices = st.get_generic_devices(name="vm", cache=False)
        d = devices[0]
        assert isinstance(d, GenericDevice)
        assert d.name == "vm"
        assert d.id == 5
        with pytest.raises(IndexError):
            devices[1]

        d = st.get_generic_devices(name="vm", cache=True)
        assert isinstance(d, GenericDevice)
        assert d.name == "vm"
        assert d.id == 5

    @responses.activate
    def test_import_generic_device(
        self, st, generic_devices_mock, sample_generic_device_csv
    ):
        res = st.import_generic_device(
            "Generic02", sample_generic_device_csv, customer_id=1
        )
        assert res is None

        res = st.import_generic_device(
            "Nonexistent", sample_generic_device_csv, customer_id=1
        )
        assert res is None

    @responses.activate
    def test_import_generic_device_error(
        self, st, generic_devices_error_mock, sample_generic_device_csv
    ):
        with pytest.raises(ValueError):
            st.import_generic_device(
                "Generic02", sample_generic_device_csv, customer_id=1
            )

        with pytest.raises(ValueError):
            st.import_generic_device(
                "Nonexistent", sample_generic_device_csv, customer_id=1
            )

    @responses.activate
    def test_delete_generic_device(self, st, generic_devices_mock):
        res = st.delete_generic_device(3)
        assert res is None

        res = st.delete_generic_device("Generic01")
        assert res is None

        with pytest.raises(ValueError):
            st.delete_generic_device("nonexistent")

    @responses.activate
    def test_delete_generic_device_error(self, st, generic_devices_error_mock):
        with pytest.raises(ValueError):
            st.delete_generic_device(3)

    @responses.activate
    def test_sync_toplogy(self, st, topology_sync_mock):
        res = st.sync_topology()
        assert res is None

    @responses.activate
    def test_toplogy_status(self, st, topology_sync_mock):
        res = st.get_topology_sync_status()
        assert isinstance(res, TopologySyncStatus)
        assert res.percentage == 100

    @responses.activate
    def test_sync_topology_auth_error(self, st, topology_sync_auth_error_mock):
        with pytest.raises(ValueError) as info:
            st.sync_topology()

        assert "Authentication" in str(info.value)

        with pytest.raises(ValueError) as info:
            st.get_topology_sync_status()

        assert "Authentication" in str(info.value)

    @responses.activate
    def test_sync_topology_500(self, st, topology_sync_500_mock):
        with pytest.raises(ValueError) as info:
            st.sync_topology()

        assert "Error synchronizing" in str(info.value)

        with pytest.raises(ValueError) as info:
            st.get_topology_sync_status()

        assert "Error getting" in str(info.value)

    @responses.activate
    def test_sync_topology_502(self, st, topology_sync_502_mock):
        with pytest.raises(HTTPError):
            st.sync_topology()

        with pytest.raises(HTTPError):
            st.get_topology_sync_status()

    @responses.activate
    def test_get_zone_subnets(self, st, zone_subnets_mock, zones_mock):
        arr1 = st.get_zone_subnets(78)
        subnet0 = arr1[0]

        assert isinstance(subnet0, ZoneEntry)

        assert subnet0.ip == IPAddress("172.16.130.0")
        assert subnet0.netmask == IPAddress("255.255.255.0")
        assert subnet0.prefix == 24
        assert subnet0.subnet == IPNetwork("172.16.130.0/24")

        arr = st.get_zone_subnets("Amsterdam_SiteA")
        subnet = arr[0]

        assert isinstance(subnet, ZoneEntry)
        assert subnet.ip == IPAddress("172.16.130.0")
        assert subnet.netmask == IPAddress("255.255.255.0")
        assert subnet.prefix == 24
        assert subnet.subnet == IPNetwork("172.16.130.0/24")
        assert subnet == subnet0

        with pytest.raises(ValueError):
            st.get_zone_subnets(200)

    @responses.activate
    def test_get_zone_descendants(self, st, zones_mock, zone_descendants_mock):
        arr = st.get_zone_descendants(80)
        zone = arr[0]
        assert isinstance(zone, ZoneReference)
        assert zone.name == "Amsterdam_Ext"
        assert zone.zones[0].name == "Amsterdam_SiteA"

        arr = st.get_zone_descendants("Amsterdam_Ext")
        zone = arr[0]
        assert isinstance(zone, ZoneReference)
        assert zone.name == "Amsterdam_Ext"
        assert zone.zones[0].name == "Amsterdam_SiteA"

        # not exist zone id
        with pytest.raises(ValueError):
            st.get_zone_descendants(200)



from dotenv import dotenv_values
from pytos2.secureapp import Sa as SecureApp
from pytos2.securechange import Scw as SecureChange
from pytos2.securetrack import St as SecureTrack
from helpers import pp, pretty, ugly, pretty_json, ugly_json


# Loads our config
config = dotenv_values(".env")

# Override or Use Config Values (Strings must be empty to use config - any value in string will be truthy)
HOST = "" or config["HOST"]
USER = "" or config["USER"]
PASS = "" or config["PASS"]

# Instantiate Classes
secure_app = SecureApp(HOST, USER, PASS)
secure_change = SecureChange(HOST, USER, PASS)
secure_track = SecureTrack(HOST, USER, PASS)



import pytest
import json
import responses
from netaddr import IPAddress

from pytos2.securetrack.license import License, Customer, SKU, SKUDevice
from pytos2.utils import get_api_node, safe_iso8601_date


class TestLicenses:
    @responses.activate
    def test_get_licenses(self, licenses_mock, st):
        license = st.get_licenses()
        assert license.supported is True
        assert license.version == "3.1"
        assert license.customer_id == "0012000001KFH3bAAH"
        assert license.customer_name == "RND License Testing"
        assert license.license_file_id == "1e54d2c3-e2b4-41e6-ad83-1153b68b17a0"
        assert license.site_name == "Eval"
        assert license.site_type.value == "production"
        assert license.installed == safe_iso8601_date("2023-10-02T03:21:02Z")
        assert license.expires == safe_iso8601_date("2024-03-05T22:00:00Z")
        assert license.status.value == "expired"
        assert license.type.value == "evaluation"

        message = license.messages[0]
        assert message.category == "LICENSE_AUDIT"
        assert message.severity == "warning"
        assert message.code == "LICENSE_USAGE_WARN_STRONG"
        assert (
            message.message
            == "License usage report must be submitted within $1 days. After this time, you will not be able to upgrade TOS."
        )
        assert message.params[0] == "27"

    @responses.activate
    def test_get_license(self, licenses_mock, st):
        license = st.get_license(license_type=1)

        assert license.id == "1"
        assert license.type == "full"
        assert license.uid == "595e0a9d-6c9e-4a42-b488-cbccb802e567"
        assert license.issued == safe_iso8601_date("2022-06-20")
        assert license.expiration == safe_iso8601_date("2024-06-17")
        assert license.customer.id == "012345678ZZZZZZXXX"
        assert license.customer.name == "Contoso"
        assert license.customer.site == "Production"
        assert license.skus[1].name == "TS-SECTRK-FW-CLS"
        assert (
            license.skus[1].description
            == "Subscribed physical, Virtual Context, or Virtual Cloud Firewall Cluster"
        )
        assert license.skus[1].quantity == 15
        assert license.skus[1].expiration == safe_iso8601_date("2024-06-18")
        assert license.skus[1].devices[0].id == "5"
        assert license.skus[1].devices[0].name == "SRX"
        assert license.skus[1].devices[0].consumed == 1

        assert license.type == "full"

        license = st.get_license(license_type="evaluation")
        assert license.type == "evaluation"

    @responses.activate
    def test_get_tiered_license(self, licenses_mock, st):
        license = st.get_tiered_license()

        assert license.supported is True
        assert license.version == "3.1"
        assert license.customer_id == "0012000001KFH3bAAH"
        assert license.customer_name == "RND License Testing"
        assert license.license_file_id == "1e54d2c3-e2b4-41e6-ad83-1153b68b17a0"
        assert license.site_name == "Eval"
        assert license.site_type.value == "production"
        assert license.installed == safe_iso8601_date("2023-10-02T03:21:02Z")
        assert license.expires == safe_iso8601_date("2024-03-05T22:00:00Z")
        assert license.status.value == "expired"
        assert license.type.value == "evaluation"

        message = license.messages[0]
        assert message.category == "LICENSE_AUDIT"
        assert message.severity == "warning"
        assert message.code == "LICENSE_USAGE_WARN_STRONG"
        assert (
            message.message
            == "License usage report must be submitted within $1 days. After this time, you will not be able to upgrade TOS."
        )
        assert message.params[0] == "27"



from datetime import datetime
from enum import Enum

from typing import List, Optional

from ..utils import propify, prop, safe_iso8601_date, kwargify
from ..models import Jsonable, Link, ObjectReference


@propify
class ApplicationConnection(Jsonable):
    services: List[ObjectReference] = prop(flatify="service", factory=list)
    sources: List[ObjectReference] = prop(flatify="source", factory=list)
    destinations: List[ObjectReference] = prop(flatify="destination", factory=list)
    status: Optional[str] = prop(None)
    external: Optional[bool] = prop(None)
    connection_to_application: Optional[ObjectReference] = prop(None)
    comment: Optional[str] = prop(None)
    uid: Optional[str] = prop(None)
    application_id: Optional[str] = prop(None, key="applicationId")
    open_tickets: List[ObjectReference] = prop(flatify="ticket", factory=list)
    name: Optional[str] = prop(None)


@propify
class Application(Jsonable):
    class Meta(Enum):
        ROOT = "application"

    id: Optional[int] = prop(None)
    comment: Optional[str] = prop(None)
    customer: Optional[ObjectReference] = prop(None)
    status: Optional[str] = prop(None)
    vendors: List[str] = prop(flatify="vendor", factory=list, jsonify=False)
    created: Optional[datetime] = prop(None, kwargify=safe_iso8601_date, jsonify=False)
    connections: List[ObjectReference] = prop(
        flatify="connection", factory=list, jsonify=False
    )
    decommissioned: Optional[bool] = prop(None)
    editors: List[ObjectReference] = prop(factory=list)
    viewers: List[ObjectReference] = prop(factory=list)
    open_tickets: List[ObjectReference] = prop(factory=list, flatify="ticket")
    connection_to_application_packs: List[ObjectReference] = prop(
        factory=list, flatify="connection_to_application_pack", jsonify=False
    )
    name: str = prop()
    owner: Optional[ObjectReference] = prop(None)
    modified: Optional[datetime] = prop(None, kwargify=safe_iso8601_date, jsonify=False)
